from test.support import import_helper, Py_GIL_DISABLED, refleak_helper
import unittest

_testcapi = import_helper.import_module('_testcapi')


class BuiltinStaticTypesTests(unittest.TestCase):

    TYPES = [
        object,
        type,
        int,
        str,
        dict,
        type(None),
        bool,
        BaseException,
        Exception,
        Warning,
        DeprecationWarning,  # Warning subclass
    ]

    def test_tp_bases_is_set(self):
        # PyTypeObject.tp_bases is documented as public API.
        # See https://github.com/python/cpython/issues/105020.
        for typeobj in self.TYPES:
            with self.subTest(typeobj):
                bases = _testcapi.type_get_tp_bases(typeobj)
                self.assertIsNot(bases, None)

    def test_tp_mro_is_set(self):
        # PyTypeObject.tp_bases is documented as public API.
        # See https://github.com/python/cpython/issues/105020.
        for typeobj in self.TYPES:
            with self.subTest(typeobj):
                mro = _testcapi.type_get_tp_mro(typeobj)
                self.assertIsNot(mro, None)


class TypeTests(unittest.TestCase):
    def test_get_type_name(self):
        class MyType:
            pass

        from _testcapi import (
            get_type_name, get_type_qualname,
            get_type_fullyqualname, get_type_module_name)

        from collections import OrderedDict
        ht = _testcapi.get_heaptype_for_name()
        for cls, fullname, modname, qualname, name in (
            (int,
             'int',
             'builtins',
             'int',
             'int'),
            (OrderedDict,
             'collections.OrderedDict',
             'collections',
             'OrderedDict',
             'OrderedDict'),
            (ht,
             '_testcapi.HeapTypeNameType',
             '_testcapi',
             'HeapTypeNameType',
             'HeapTypeNameType'),
            (MyType,
             f'{__name__}.TypeTests.test_get_type_name.<locals>.MyType',
             __name__,
             'TypeTests.test_get_type_name.<locals>.MyType',
             'MyType'),
        ):
            with self.subTest(cls=repr(cls)):
                self.assertEqual(get_type_fullyqualname(cls), fullname)
                self.assertEqual(get_type_module_name(cls), modname)
                self.assertEqual(get_type_qualname(cls), qualname)
                self.assertEqual(get_type_name(cls), name)

        # override __module__
        ht.__module__ = 'test_module'
        self.assertEqual(get_type_fullyqualname(ht), 'test_module.HeapTypeNameType')
        self.assertEqual(get_type_module_name(ht), 'test_module')
        self.assertEqual(get_type_qualname(ht), 'HeapTypeNameType')
        self.assertEqual(get_type_name(ht), 'HeapTypeNameType')

        # override __name__ and __qualname__
        MyType.__name__ = 'my_name'
        MyType.__qualname__ = 'my_qualname'
        self.assertEqual(get_type_fullyqualname(MyType), f'{__name__}.my_qualname')
        self.assertEqual(get_type_module_name(MyType), __name__)
        self.assertEqual(get_type_qualname(MyType), 'my_qualname')
        self.assertEqual(get_type_name(MyType), 'my_name')

        # override also __module__
        MyType.__module__ = 'my_module'
        self.assertEqual(get_type_fullyqualname(MyType), 'my_module.my_qualname')
        self.assertEqual(get_type_module_name(MyType), 'my_module')
        self.assertEqual(get_type_qualname(MyType), 'my_qualname')
        self.assertEqual(get_type_name(MyType), 'my_name')

        # PyType_GetFullyQualifiedName() ignores the module if it's "builtins"
        # or "__main__" of it is not a string
        MyType.__module__ = 'builtins'
        self.assertEqual(get_type_fullyqualname(MyType), 'my_qualname')
        MyType.__module__ = '__main__'
        self.assertEqual(get_type_fullyqualname(MyType), 'my_qualname')
        MyType.__module__ = 123
        self.assertEqual(get_type_fullyqualname(MyType), 'my_qualname')

    def test_get_base_by_token(self):
        def get_base_by_token(src, key, comparable=True):
            def run(use_mro):
                find_first = _testcapi.pytype_getbasebytoken
                ret1, result = find_first(src, key, use_mro, True)
                ret2, no_result = find_first(src, key, use_mro, False)
                self.assertIn(ret1, (0, 1))
                self.assertEqual(ret1, result is not None)
                self.assertEqual(ret1, ret2)
                self.assertIsNone(no_result)
                return result

            found_in_mro = run(True)
            found_in_bases = run(False)
            if comparable:
                self.assertIs(found_in_mro, found_in_bases)
                return found_in_mro
            return found_in_mro, found_in_bases

        create_type = _testcapi.create_type_with_token
        get_token = _testcapi.get_tp_token

        Py_TP_USE_SPEC = _testcapi.Py_TP_USE_SPEC
        self.assertEqual(Py_TP_USE_SPEC, 0)

        A1 = create_type('_testcapi.A1', Py_TP_USE_SPEC)
        self.assertTrue(get_token(A1) != Py_TP_USE_SPEC)

        B1 = create_type('_testcapi.B1', id(self))
        self.assertTrue(get_token(B1) == id(self))

        tokenA1 = get_token(A1)
        # find A1 from A1
        found = get_base_by_token(A1, tokenA1)
        self.assertIs(found, A1)

        # no token in static types
        STATIC = type(1)
        self.assertEqual(get_token(STATIC), 0)
        found = get_base_by_token(STATIC, tokenA1)
        self.assertIs(found, None)

        # no token in pure subtypes
        class A2(A1): pass
        self.assertEqual(get_token(A2), 0)
        # find A1
        class Z(STATIC, B1, A2): pass
        found = get_base_by_token(Z, tokenA1)
        self.assertIs(found, A1)

        # searching for NULL token is an error
        with self.assertRaises(SystemError):
            get_base_by_token(Z, 0)
        with self.assertRaises(SystemError):
            get_base_by_token(STATIC, 0)

        # share the token with A1
        C1 = create_type('_testcapi.C1', tokenA1)
        self.assertTrue(get_token(C1) == tokenA1)

        # find C1 first by shared token
        class Z(C1, A2): pass
        found = get_base_by_token(Z, tokenA1)
        self.assertIs(found, C1)
        # B1 not found
        found = get_base_by_token(Z, get_token(B1))
        self.assertIs(found, None)

        with self.assertRaises(TypeError):
            _testcapi.pytype_getbasebytoken(
                'not a type', id(self), True, False)

    def test_get_module_by_def(self):
        heaptype = _testcapi.create_type_with_token('_testcapi.H', 0)
        mod = _testcapi.pytype_getmodulebydef(heaptype)
        self.assertIs(mod, _testcapi)

        class H1(heaptype): pass
        mod = _testcapi.pytype_getmodulebydef(H1)
        self.assertIs(mod, _testcapi)

        with self.assertRaises(TypeError):
            _testcapi.pytype_getmodulebydef(int)

        class H2(int): pass
        with self.assertRaises(TypeError):
            _testcapi.pytype_getmodulebydef(H2)

    def test_freeze(self):
        # test PyType_Freeze()
        type_freeze = _testcapi.type_freeze

        # simple case, no inherante
        class MyType:
            pass
        MyType.attr = "mutable"

        type_freeze(MyType)
        err_msg = "cannot set 'attr' attribute of immutable type 'MyType'"
        with self.assertRaisesRegex(TypeError, err_msg):
            # the class is now immutable
            MyType.attr = "immutable"

        # test MRO: PyType_Freeze() requires base classes to be immutable
        class A: pass
        class B: pass
        class C(B): pass
        class D(A, C): pass

        self.assertEqual(D.mro(), [D, A, C, B, object])
        with self.assertRaises(TypeError):
            type_freeze(D)

        type_freeze(A)
        type_freeze(B)
        type_freeze(C)
        # all parent classes are now immutable, so D can be made immutable
        # as well
        type_freeze(D)

    @unittest.skipIf(
        Py_GIL_DISABLED and refleak_helper.hunting_for_refleaks(),
        "Specialization failure triggers gh-127773")
    def test_freeze_meta(self):
        """test PyType_Freeze() with overridden MRO"""
        type_freeze = _testcapi.type_freeze

        class Base:
            value = 1

        class Meta(type):
            def mro(cls):
                return (cls, Base, object)

        class FreezeThis(metaclass=Meta):
            """This has `Base` in the MRO, but not tp_bases"""

        self.assertEqual(FreezeThis.value, 1)

        with self.assertRaises(TypeError):
            type_freeze(FreezeThis)

        Base.value = 2
        self.assertEqual(FreezeThis.value, 2)

        type_freeze(Base)
        with self.assertRaises(TypeError):
            Base.value = 3
        type_freeze(FreezeThis)
        self.assertEqual(FreezeThis.value, 2)

    def test_manual_heap_type(self):
        # gh-128923: test that a manually allocated and initailized heap type
        # works correctly
        ManualHeapType = _testcapi.ManualHeapType
        for i in range(100):
            self.assertIsInstance(ManualHeapType(), ManualHeapType)

    def test_extension_managed_dict_type(self):
        ManagedDictType = _testcapi.ManagedDictType
        obj = ManagedDictType()
        obj.foo = 42
        self.assertEqual(obj.foo, 42)
        self.assertEqual(obj.__dict__, {'foo': 42})
        obj.__dict__ = {'bar': 3}
        self.assertEqual(obj.__dict__, {'bar': 3})
        self.assertEqual(obj.bar, 3)



import dis
import gc
from itertools import combinations, product
import opcode
import sys
import textwrap
import unittest
try:
    import _testinternalcapi
except ImportError:
    _testinternalcapi = None

from test import support
from test.support.bytecode_helper import (
    BytecodeTestCase, CfgOptimizationTestCase, CompilationStepTestCase)


def compile_pattern_with_fast_locals(pattern):
    source = textwrap.dedent(
        f"""
        def f(x):
            match x:
                case {pattern}:
                    pass
        """
    )
    namespace = {}
    exec(source, namespace)
    return namespace["f"].__code__


def count_instr_recursively(f, opname):
    count = 0
    for instr in dis.get_instructions(f):
        if instr.opname == opname:
            count += 1
    if hasattr(f, '__code__'):
        f = f.__code__
    for c in f.co_consts:
        if hasattr(c, 'co_code'):
            count += count_instr_recursively(c, opname)
    return count


def get_binop_argval(arg):
    for i, nb_op in enumerate(opcode._nb_ops):
        if arg == nb_op[0]:
            return i
    assert False, f"{arg} is not a valid BINARY_OP argument."


class TestTranforms(BytecodeTestCase):

    def check_jump_targets(self, code):
        instructions = list(dis.get_instructions(code))
        targets = {instr.offset: instr for instr in instructions}
        for instr in instructions:
            if 'JUMP_' not in instr.opname:
                continue
            tgt = targets[instr.argval]
            # jump to unconditional jump
            if tgt.opname in ('JUMP_BACKWARD', 'JUMP_FORWARD'):
                self.fail(f'{instr.opname} at {instr.offset} '
                          f'jumps to {tgt.opname} at {tgt.offset}')
            # unconditional jump to RETURN_VALUE
            if (instr.opname in ('JUMP_BACKWARD', 'JUMP_FORWARD') and
                tgt.opname == 'RETURN_VALUE'):
                self.fail(f'{instr.opname} at {instr.offset} '
                          f'jumps to {tgt.opname} at {tgt.offset}')

    def check_lnotab(self, code):
        "Check that the lnotab byte offsets are sensible."
        code = dis._get_code_object(code)
        lnotab = list(dis.findlinestarts(code))
        # Don't bother checking if the line info is sensible, because
        # most of the line info we can get at comes from lnotab.
        min_bytecode = min(t[0] for t in lnotab)
        max_bytecode = max(t[0] for t in lnotab)
        self.assertGreaterEqual(min_bytecode, 0)
        self.assertLess(max_bytecode, len(code.co_code))
        # This could conceivably test more (and probably should, as there
        # aren't very many tests of lnotab), if peepholer wasn't scheduled
        # to be replaced anyway.

    def test_unot(self):
        # UNARY_NOT POP_JUMP_IF_FALSE  -->  POP_JUMP_IF_TRUE'
        def unot(x):
            if not x == 2:
                del x
        self.assertNotInBytecode(unot, 'UNARY_NOT')
        self.assertNotInBytecode(unot, 'POP_JUMP_IF_FALSE')
        self.assertInBytecode(unot, 'POP_JUMP_IF_TRUE')
        self.check_lnotab(unot)

    def test_elim_inversion_of_is_or_in(self):
        for line, cmp_op, invert in (
            ('not a is b', 'IS_OP', 1,),
            ('not a is not b', 'IS_OP', 0,),
            ('not a in b', 'CONTAINS_OP', 1,),
            ('not a not in b', 'CONTAINS_OP', 0,),
            ):
            with self.subTest(line=line):
                code = compile(line, '', 'single')
                self.assertInBytecode(code, cmp_op, invert)
                self.check_lnotab(code)

    def test_global_as_constant(self):
        # LOAD_GLOBAL None/True/False  -->  LOAD_CONST None/True/False
        def f():
            x = None
            x = None
            return x
        def g():
            x = True
            return x
        def h():
            x = False
            return x

        for func, elem in ((f, None), (g, True), (h, False)):
            with self.subTest(func=func):
                self.assertNotInBytecode(func, 'LOAD_GLOBAL')
                self.assertInBytecode(func, 'LOAD_CONST', elem)
                self.check_lnotab(func)

        def f():
            'Adding a docstring made this test fail in Py2.5.0'
            return None

        self.assertNotInBytecode(f, 'LOAD_GLOBAL')
        self.assertInBytecode(f, 'LOAD_CONST', None)
        self.check_lnotab(f)

    def test_while_one(self):
        # Skip over:  LOAD_CONST trueconst  POP_JUMP_IF_FALSE xx
        def f():
            while 1:
                pass
            return list
        for elem in ('LOAD_CONST', 'POP_JUMP_IF_FALSE'):
            self.assertNotInBytecode(f, elem)
        for elem in ('JUMP_BACKWARD',):
            self.assertInBytecode(f, elem)
        self.check_lnotab(f)

    def test_pack_unpack(self):
        for line, elem in (
            ('a, = a,', 'LOAD_CONST',),
            ('a, b = a, b', 'SWAP',),
            ('a, b, c = a, b, c', 'SWAP',),
            ):
            with self.subTest(line=line):
                code = compile(line,'','single')
                self.assertInBytecode(code, elem)
                self.assertNotInBytecode(code, 'BUILD_TUPLE')
                self.assertNotInBytecode(code, 'UNPACK_SEQUENCE')
                self.check_lnotab(code)

    def test_constant_folding_tuples_of_constants(self):
        for line, elem in (
            ('a = 1,2,3', (1, 2, 3)),
            ('("a","b","c")', ('a', 'b', 'c')),
            ('a,b,c,d = 1,2,3,4', (1, 2, 3, 4)),
            ('(None, 1, None)', (None, 1, None)),
            ('((1, 2), 3, 4)', ((1, 2), 3, 4)),
            ):
            with self.subTest(line=line):
                code = compile(line,'','single')
                self.assertInBytecode(code, 'LOAD_CONST', elem)
                self.assertNotInBytecode(code, 'BUILD_TUPLE')
                self.check_lnotab(code)

        # Long tuples should be folded too.
        code = compile(repr(tuple(range(10000))),'','single')
        self.assertNotInBytecode(code, 'BUILD_TUPLE')
        # One LOAD_CONST for the tuple, one for the None return value
        load_consts = [instr for instr in dis.get_instructions(code)
                              if instr.opname == 'LOAD_CONST']
        self.assertEqual(len(load_consts), 2)
        self.check_lnotab(code)

        # Bug 1053819:  Tuple of constants misidentified when presented with:
        # . . . opcode_with_arg 100   unary_opcode   BUILD_TUPLE 1  . . .
        # The following would segfault upon compilation
        def crater():
            (~[
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
                0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
            ],)
        self.check_lnotab(crater)

    def test_constant_folding_lists_of_constants(self):
        for line, elem in (
            # in/not in constants with BUILD_LIST should be folded to a tuple:
            ('a in [1,2,3]', (1, 2, 3)),
            ('a not in ["a","b","c"]', ('a', 'b', 'c')),
            ('a in [None, 1, None]', (None, 1, None)),
            ('a not in [(1, 2), 3, 4]', ((1, 2), 3, 4)),
            ):
            with self.subTest(line=line):
                code = compile(line, '', 'single')
                self.assertInBytecode(code, 'LOAD_CONST', elem)
                self.assertNotInBytecode(code, 'BUILD_LIST')
                self.check_lnotab(code)

    def test_constant_folding_sets_of_constants(self):
        for line, elem in (
            # in/not in constants with BUILD_SET should be folded to a frozenset:
            ('a in {1,2,3}', frozenset({1, 2, 3})),
            ('a not in {"a","b","c"}', frozenset({'a', 'c', 'b'})),
            ('a in {None, 1, None}', frozenset({1, None})),
            ('a not in {(1, 2), 3, 4}', frozenset({(1, 2), 3, 4})),
            ('a in {1, 2, 3, 3, 2, 1}', frozenset({1, 2, 3})),
            ):
            with self.subTest(line=line):
                code = compile(line, '', 'single')
                self.assertNotInBytecode(code, 'BUILD_SET')
                self.assertInBytecode(code, 'LOAD_CONST', elem)
                self.check_lnotab(code)

        # Ensure that the resulting code actually works:
        def f(a):
            return a in {1, 2, 3}

        def g(a):
            return a not in {1, 2, 3}

        self.assertTrue(f(3))
        self.assertTrue(not f(4))
        self.check_lnotab(f)

        self.assertTrue(not g(3))
        self.assertTrue(g(4))
        self.check_lnotab(g)

    def test_constant_folding_small_int(self):
        tests = [
            ('(0, )[0]', 0),
            ('(1 + 2, )[0]', 3),
            ('(2 + 2 * 2, )[0]', 6),
            ('(1, (1 + 2 + 3, ))[1][0]', 6),
            ('1 + 2', 3),
            ('2 + 2 * 2 // 2 - 2', 2),
            ('(255, )[0]', 255),
            ('(256, )[0]', None),
            ('(1000, )[0]', None),
            ('(1 - 2, )[0]', None),
            ('255 + 0', 255),
            ('255 + 1', None),
            ('-1', None),
            ('--1', 1),
            ('--255', 255),
            ('--256', None),
            ('~1', None),
            ('~~1', 1),
            ('~~255', 255),
            ('~~256', None),
            ('++255', 255),
            ('++256', None),
        ]
        for expr, oparg in tests:
            with self.subTest(expr=expr, oparg=oparg):
                code = compile(expr, '', 'single')
                if oparg is not None:
                    self.assertInBytecode(code, 'LOAD_SMALL_INT', oparg)
                else:
                    self.assertNotInBytecode(code, 'LOAD_SMALL_INT')
                self.check_lnotab(code)

    def test_constant_folding_unaryop(self):
        intrinsic_positive = 5
        tests = [
            ('-0', 'UNARY_NEGATIVE', None, True, 'LOAD_SMALL_INT', 0),
            ('-0.0', 'UNARY_NEGATIVE', None, True, 'LOAD_CONST', -0.0),
            ('-(1.0-1.0)', 'UNARY_NEGATIVE', None, True, 'LOAD_CONST', -0.0),
            ('-0.5', 'UNARY_NEGATIVE', None, True, 'LOAD_CONST', -0.5),
            ('---1', 'UNARY_NEGATIVE', None, True, 'LOAD_CONST', -1),
            ('---""', 'UNARY_NEGATIVE', None, False, None, None),
            ('~~~1', 'UNARY_INVERT', None, True, 'LOAD_CONST', -2),
            ('~~~""', 'UNARY_INVERT', None, False, None, None),
            ('not not True', 'UNARY_NOT', None, True, 'LOAD_CONST', True),
            ('not not x', 'UNARY_NOT', None, True, 'LOAD_NAME', 'x'),  # this should be optimized regardless of constant or not
            ('+++1', 'CALL_INTRINSIC_1', intrinsic_positive, True, 'LOAD_SMALL_INT', 1),
            ('---x', 'UNARY_NEGATIVE', None, False, None, None),
            ('~~~x', 'UNARY_INVERT', None, False, None, None),
            ('+++x', 'CALL_INTRINSIC_1', intrinsic_positive, False, None, None),
            ('~True', 'UNARY_INVERT', None, False, None, None),
        ]

        for (
            expr,
            original_opcode,
            original_argval,
            is_optimized,
            optimized_opcode,
            optimized_argval,
        ) in tests:
            with self.subTest(expr=expr, is_optimized=is_optimized):
                code = compile(expr, "", "single")
                if is_optimized:
                    self.assertNotInBytecode(code, original_opcode, argval=original_argval)
                    self.assertInBytecode(code, optimized_opcode, argval=optimized_argval)
                else:
                    self.assertInBytecode(code, original_opcode, argval=original_argval)
                self.check_lnotab(code)

        # Check that -0.0 works after marshaling
        def negzero():
            return -(1.0-1.0)

        for instr in dis.get_instructions(negzero):
            self.assertNotStartsWith(instr.opname, 'UNARY_')
        self.check_lnotab(negzero)

    def test_constant_folding_binop(self):
        tests = [
            ('1 + 2', 'NB_ADD', True, 'LOAD_SMALL_INT', 3),
            ('1 + 2 + 3', 'NB_ADD', True, 'LOAD_SMALL_INT', 6),
            ('1 + ""', 'NB_ADD', False, None, None),
            ('1 - 2', 'NB_SUBTRACT', True, 'LOAD_CONST', -1),
            ('1 - 2 - 3', 'NB_SUBTRACT', True, 'LOAD_CONST', -4),
            ('1 - ""', 'NB_SUBTRACT', False, None, None),
            ('2 * 2', 'NB_MULTIPLY', True, 'LOAD_SMALL_INT', 4),
            ('2 * 2 * 2', 'NB_MULTIPLY', True, 'LOAD_SMALL_INT', 8),
            ('2 / 2', 'NB_TRUE_DIVIDE', True, 'LOAD_CONST', 1.0),
            ('2 / 2 / 2', 'NB_TRUE_DIVIDE', True, 'LOAD_CONST', 0.5),
            ('2 / ""', 'NB_TRUE_DIVIDE', False, None, None),
            ('2 // 2', 'NB_FLOOR_DIVIDE', True, 'LOAD_SMALL_INT', 1),
            ('2 // 2 // 2', 'NB_FLOOR_DIVIDE', True, 'LOAD_SMALL_INT', 0),
            ('2 // ""', 'NB_FLOOR_DIVIDE', False, None, None),
            ('2 % 2', 'NB_REMAINDER', True, 'LOAD_SMALL_INT', 0),
            ('2 % 2 % 2', 'NB_REMAINDER', True, 'LOAD_SMALL_INT', 0),
            ('2 % ()', 'NB_REMAINDER', False, None, None),
            ('2 ** 2', 'NB_POWER', True, 'LOAD_SMALL_INT', 4),
            ('2 ** 2 ** 2', 'NB_POWER', True, 'LOAD_SMALL_INT', 16),
            ('2 ** ""', 'NB_POWER', False, None, None),
            ('2 << 2', 'NB_LSHIFT', True, 'LOAD_SMALL_INT', 8),
            ('2 << 2 << 2', 'NB_LSHIFT', True, 'LOAD_SMALL_INT', 32),
            ('2 << ""', 'NB_LSHIFT', False, None, None),
            ('2 >> 2', 'NB_RSHIFT', True, 'LOAD_SMALL_INT', 0),
            ('2 >> 2 >> 2', 'NB_RSHIFT', True, 'LOAD_SMALL_INT', 0),
            ('2 >> ""', 'NB_RSHIFT', False, None, None),
            ('2 | 2', 'NB_OR', True, 'LOAD_SMALL_INT', 2),
            ('2 | 2 | 2', 'NB_OR', True, 'LOAD_SMALL_INT', 2),
            ('2 | ""', 'NB_OR', False, None, None),
            ('2 & 2', 'NB_AND', True, 'LOAD_SMALL_INT', 2),
            ('2 & 2 & 2', 'NB_AND', True, 'LOAD_SMALL_INT', 2),
            ('2 & ""', 'NB_AND', False, None, None),
            ('2 ^ 2', 'NB_XOR', True, 'LOAD_SMALL_INT', 0),
            ('2 ^ 2 ^ 2', 'NB_XOR', True, 'LOAD_SMALL_INT', 2),
            ('2 ^ ""', 'NB_XOR', False, None, None),
            ('(1, )[0]', 'NB_SUBSCR', True, 'LOAD_SMALL_INT', 1),
            ('(1, )[-1]', 'NB_SUBSCR', True, 'LOAD_SMALL_INT', 1),
            ('(1 + 2, )[0]', 'NB_SUBSCR', True, 'LOAD_SMALL_INT', 3),
            ('(1, (1, 2))[1][1]', 'NB_SUBSCR', True, 'LOAD_SMALL_INT', 2),
            ('(1, 2)[2-1]', 'NB_SUBSCR', True, 'LOAD_SMALL_INT', 2),
            ('(1, (1, 2))[1][2-1]', 'NB_SUBSCR', True, 'LOAD_SMALL_INT', 2),
            ('(1, (1, 2))[1:6][0][2-1]', 'NB_SUBSCR', True, 'LOAD_SMALL_INT', 2),
            ('"a"[0]', 'NB_SUBSCR', True, 'LOAD_CONST', 'a'),
            ('("a" + "b")[1]', 'NB_SUBSCR', True, 'LOAD_CONST', 'b'),
            ('("a" + "b", )[0][1]', 'NB_SUBSCR', True, 'LOAD_CONST', 'b'),
            ('("a" * 10)[9]', 'NB_SUBSCR', True, 'LOAD_CONST', 'a'),
            ('(1, )[1]', 'NB_SUBSCR', False, None, None),
            ('(1, )[-2]', 'NB_SUBSCR', False, None, None),
            ('"a"[1]', 'NB_SUBSCR', False, None, None),
            ('"a"[-2]', 'NB_SUBSCR', False, None, None),
            ('("a" + "b")[2]', 'NB_SUBSCR', False, None, None),
            ('("a" + "b", )[0][2]', 'NB_SUBSCR', False, None, None),
            ('("a" + "b", )[1][0]', 'NB_SUBSCR', False, None, None),
            ('("a" * 10)[10]', 'NB_SUBSCR', False, None, None),
            ('(1, (1, 2))[2:6][0][2-1]', 'NB_SUBSCR', False, None, None),
        ]

        for (
            expr,
            nb_op,
            is_optimized,
            optimized_opcode,
            optimized_argval
        ) in tests:
            with self.subTest(expr=expr, is_optimized=is_optimized):
                code = compile(expr, '', 'single')
                nb_op_val = get_binop_argval(nb_op)
                if is_optimized:
                    self.assertNotInBytecode(code, 'BINARY_OP', argval=nb_op_val)
                    self.assertInBytecode(code, optimized_opcode, argval=optimized_argval)
                else:
                    self.assertInBytecode(code, 'BINARY_OP', argval=nb_op_val)
                self.check_lnotab(code)

        # Verify that large sequences do not result from folding
        code = compile('"x"*10000', '', 'single')
        self.assertInBytecode(code, 'LOAD_CONST', 10000)
        self.assertNotIn("x"*10000, code.co_consts)
        self.check_lnotab(code)
        code = compile('1<<1000', '', 'single')
        self.assertInBytecode(code, 'LOAD_CONST', 1000)
        self.assertNotIn(1<<1000, code.co_consts)
        self.check_lnotab(code)
        code = compile('2**1000', '', 'single')
        self.assertInBytecode(code, 'LOAD_CONST', 1000)
        self.assertNotIn(2**1000, code.co_consts)
        self.check_lnotab(code)

        # Test binary subscript on unicode
        # valid code get optimized
        code = compile('"foo"[0]', '', 'single')
        self.assertInBytecode(code, 'LOAD_CONST', 'f')
        self.assertNotInBytecode(code, 'BINARY_OP')
        self.check_lnotab(code)
        code = compile('"\u0061\uffff"[1]', '', 'single')
        self.assertInBytecode(code, 'LOAD_CONST', '\uffff')
        self.assertNotInBytecode(code,'BINARY_OP')
        self.check_lnotab(code)

        # With PEP 393, non-BMP char get optimized
        code = compile('"\U00012345"[0]', '', 'single')
        self.assertInBytecode(code, 'LOAD_CONST', '\U00012345')
        self.assertNotInBytecode(code, 'BINARY_OP')
        self.check_lnotab(code)

        # invalid code doesn't get optimized
        # out of range
        code = compile('"fuu"[10]', '', 'single')
        self.assertInBytecode(code, 'BINARY_OP')
        self.check_lnotab(code)


    def test_constant_folding_remove_nop_location(self):
        sources = [
            """
            (-
             -
             -
             1)
            """,

            """
            (1
             +
             2
             +
             3)
            """,

            """
            (1,
             2,
             3)[0]
            """,

            """
            [1,
             2,
             3]
            """,

            """
            {1,
             2,
             3}
            """,

            """
            1 in [
               1,
               2,
               3
            ]
            """,

            """
            1 in {
               1,
               2,
               3
            }
            """,

            """
            for _ in [1,
                      2,
                      3]:
                pass
            """,

            """
            for _ in [1,
                      2,
                      x]:
                pass
            """,

            """
            for _ in {1,
                      2,
                      3}:
                pass
            """
        ]

        for source in sources:
            code = compile(textwrap.dedent(source), '', 'single')
            self.assertNotInBytecode(code, 'NOP')

    def test_elim_extra_return(self):
        # RETURN LOAD_CONST None RETURN  -->  RETURN
        def f(x):
            return x
        self.assertNotInBytecode(f, 'LOAD_CONST', None)
        returns = [instr for instr in dis.get_instructions(f)
                          if instr.opname == 'RETURN_VALUE']
        self.assertEqual(len(returns), 1)
        self.check_lnotab(f)

    def test_elim_jump_to_return(self):
        # JUMP_FORWARD to RETURN -->  RETURN
        def f(cond, true_value, false_value):
            # Intentionally use two-line expression to test issue37213.
            return (true_value if cond
                    else false_value)
        self.check_jump_targets(f)
        self.assertNotInBytecode(f, 'JUMP_FORWARD')
        self.assertNotInBytecode(f, 'JUMP_BACKWARD')
        returns = [instr for instr in dis.get_instructions(f)
                          if instr.opname == 'RETURN_VALUE']
        self.assertEqual(len(returns), 2)
        self.check_lnotab(f)

    def test_elim_jump_to_uncond_jump(self):
        # POP_JUMP_IF_FALSE to JUMP_FORWARD --> POP_JUMP_IF_FALSE to non-jump
        def f():
            if a:
                # Intentionally use two-line expression to test issue37213.
                if (c
                    or d):
                    foo()
            else:
                baz()
        self.check_jump_targets(f)
        self.check_lnotab(f)

    def test_elim_jump_to_uncond_jump2(self):
        # POP_JUMP_IF_FALSE to JUMP_BACKWARD --> POP_JUMP_IF_FALSE to non-jump
        def f():
            while a:
                # Intentionally use two-line expression to test issue37213.
                if (c
                    or d):
                    a = foo()
        self.check_jump_targets(f)
        self.check_lnotab(f)

    def test_elim_jump_to_uncond_jump3(self):
        # Intentionally use two-line expressions to test issue37213.
        # POP_JUMP_IF_FALSE to POP_JUMP_IF_FALSE --> POP_JUMP_IF_FALSE to non-jump
        def f(a, b, c):
            return ((a and b)
                    and c)
        self.check_jump_targets(f)
        self.check_lnotab(f)
        self.assertEqual(count_instr_recursively(f, 'POP_JUMP_IF_FALSE'), 2)
        # POP_JUMP_IF_TRUE to POP_JUMP_IF_TRUE --> POP_JUMP_IF_TRUE to non-jump
        def f(a, b, c):
            return ((a or b)
                    or c)
        self.check_jump_targets(f)
        self.check_lnotab(f)
        self.assertEqual(count_instr_recursively(f, 'POP_JUMP_IF_TRUE'), 2)
        # JUMP_IF_FALSE_OR_POP to JUMP_IF_TRUE_OR_POP --> POP_JUMP_IF_FALSE to non-jump
        def f(a, b, c):
            return ((a and b)
                    or c)
        self.check_jump_targets(f)
        self.check_lnotab(f)
        self.assertEqual(count_instr_recursively(f, 'POP_JUMP_IF_FALSE'), 1)
        self.assertEqual(count_instr_recursively(f, 'POP_JUMP_IF_TRUE'), 1)
        # POP_JUMP_IF_TRUE to POP_JUMP_IF_FALSE --> POP_JUMP_IF_TRUE to non-jump
        def f(a, b, c):
            return ((a or b)
                    and c)
        self.check_jump_targets(f)
        self.check_lnotab(f)
        self.assertEqual(count_instr_recursively(f, 'POP_JUMP_IF_FALSE'), 1)
        self.assertEqual(count_instr_recursively(f, 'POP_JUMP_IF_TRUE'), 1)

    def test_elim_jump_to_uncond_jump4(self):
        def f():
            for i in range(5):
                if i > 3:
                    print(i)
        self.check_jump_targets(f)

    def test_elim_jump_after_return1(self):
        # Eliminate dead code: jumps immediately after returns can't be reached
        def f(cond1, cond2):
            if cond1: return 1
            if cond2: return 2
            while 1:
                return 3
            while 1:
                if cond1: return 4
                return 5
            return 6
        self.assertNotInBytecode(f, 'JUMP_FORWARD')
        self.assertNotInBytecode(f, 'JUMP_BACKWARD')
        returns = [instr for instr in dis.get_instructions(f)
                          if instr.opname == 'RETURN_VALUE']
        self.assertLessEqual(len(returns), 6)
        self.check_lnotab(f)

    def test_make_function_doesnt_bail(self):
        def f():
            def g()->1+1:
                pass
            return g
        self.assertNotInBytecode(f, 'BINARY_OP')
        self.check_lnotab(f)

    def test_in_literal_list(self):
        def containtest():
            return x in [a, b]
        self.assertEqual(count_instr_recursively(containtest, 'BUILD_LIST'), 0)
        self.check_lnotab(containtest)

    def test_iterate_literal_list(self):
        def forloop():
            for x in [a, b]:
                pass
        self.assertEqual(count_instr_recursively(forloop, 'BUILD_LIST'), 0)
        self.check_lnotab(forloop)

    def test_condition_with_binop_with_bools(self):
        def f():
            if True or False:
                return 1
            return 0
        self.assertEqual(f(), 1)
        self.check_lnotab(f)

    def test_if_with_if_expression(self):
        # Check bpo-37289
        def f(x):
            if (True if x else False):
                return True
            return False
        self.assertTrue(f(True))
        self.check_lnotab(f)

    def test_trailing_nops(self):
        # Check the lnotab of a function that even after trivial
        # optimization has trailing nops, which the lnotab adjustment has to
        # handle properly (bpo-38115).
        def f(x):
            while 1:
                return 3
            while 1:
                return 5
            return 6
        self.check_lnotab(f)

    def test_assignment_idiom_in_comprehensions(self):
        def listcomp():
            return [y for x in a for y in [f(x)]]
        self.assertEqual(count_instr_recursively(listcomp, 'FOR_ITER'), 1)
        def setcomp():
            return {y for x in a for y in [f(x)]}
        self.assertEqual(count_instr_recursively(setcomp, 'FOR_ITER'), 1)
        def dictcomp():
            return {y: y for x in a for y in [f(x)]}
        self.assertEqual(count_instr_recursively(dictcomp, 'FOR_ITER'), 1)
        def genexpr():
            return (y for x in a for y in [f(x)])
        self.assertEqual(count_instr_recursively(genexpr, 'FOR_ITER'), 1)

    @support.requires_resource('cpu')
    def test_format_combinations(self):
        flags = '-+ #0'
        testcases = [
            *product(('', '1234', 'абвг'), 'sra'),
            *product((1234, -1234), 'duioxX'),
            *product((1234.5678901, -1234.5678901), 'duifegFEG'),
            *product((float('inf'), -float('inf')), 'fegFEG'),
        ]
        width_precs = [
            *product(('', '1', '30'), ('', '.', '.0', '.2')),
            ('', '.40'),
            ('30', '.40'),
        ]
        for value, suffix in testcases:
            for width, prec in width_precs:
                for r in range(len(flags) + 1):
                    for spec in combinations(flags, r):
                        fmt = '%' + ''.join(spec) + width + prec + suffix
                        with self.subTest(fmt=fmt, value=value):
                            s1 = fmt % value
                            s2 = eval(f'{fmt!r} % (x,)', {'x': value})
                            self.assertEqual(s2, s1, f'{fmt = }')

    def test_format_misc(self):
        def format(fmt, *values):
            vars = [f'x{i+1}' for i in range(len(values))]
            if len(vars) == 1:
                args = '(' + vars[0] + ',)'
            else:
                args = '(' + ', '.join(vars) + ')'
            return eval(f'{fmt!r} % {args}', dict(zip(vars, values)))

        self.assertEqual(format('string'), 'string')
        self.assertEqual(format('x = %s!', 1234), 'x = 1234!')
        self.assertEqual(format('x = %d!', 1234), 'x = 1234!')
        self.assertEqual(format('x = %x!', 1234), 'x = 4d2!')
        self.assertEqual(format('x = %f!', 1234), 'x = 1234.000000!')
        self.assertEqual(format('x = %s!', 1234.0000625), 'x = 1234.0000625!')
        self.assertEqual(format('x = %f!', 1234.0000625), 'x = 1234.000063!')
        self.assertEqual(format('x = %d!', 1234.0000625), 'x = 1234!')
        self.assertEqual(format('x = %s%% %%%%', 1234), 'x = 1234% %%')
        self.assertEqual(format('x = %s!', '%% %s'), 'x = %% %s!')
        self.assertEqual(format('x = %s, y = %d', 12, 34), 'x = 12, y = 34')

    def test_format_errors(self):
        with self.assertRaisesRegex(TypeError,
                    'not enough arguments for format string'):
            eval("'%s' % ()")
        with self.assertRaisesRegex(TypeError,
                    'not all arguments converted during string formatting'):
            eval("'%s' % (x, y)", {'x': 1, 'y': 2})
        with self.assertRaisesRegex(ValueError, 'incomplete format'):
            eval("'%s%' % (x,)", {'x': 1234})
        with self.assertRaisesRegex(ValueError, 'incomplete format'):
            eval("'%s%%%' % (x,)", {'x': 1234})
        with self.assertRaisesRegex(TypeError,
                    'not enough arguments for format string'):
            eval("'%s%z' % (x,)", {'x': 1234})
        with self.assertRaisesRegex(ValueError, 'unsupported format character'):
            eval("'%s%z' % (x, 5)", {'x': 1234})
        with self.assertRaisesRegex(TypeError, 'a real number is required, not str'):
            eval("'%d' % (x,)", {'x': '1234'})
        with self.assertRaisesRegex(TypeError, 'an integer is required, not float'):
            eval("'%x' % (x,)", {'x': 1234.56})
        with self.assertRaisesRegex(TypeError, 'an integer is required, not str'):
            eval("'%x' % (x,)", {'x': '1234'})
        with self.assertRaisesRegex(TypeError, 'must be real number, not str'):
            eval("'%f' % (x,)", {'x': '1234'})
        with self.assertRaisesRegex(TypeError,
                    'not enough arguments for format string'):
            eval("'%s, %s' % (x, *y)", {'x': 1, 'y': []})
        with self.assertRaisesRegex(TypeError,
                    'not all arguments converted during string formatting'):
            eval("'%s, %s' % (x, *y)", {'x': 1, 'y': [2, 3]})

    def test_static_swaps_unpack_two(self):
        def f(a, b):
            a, b = a, b
            b, a = a, b
        self.assertNotInBytecode(f, "SWAP")

    def test_static_swaps_unpack_three(self):
        def f(a, b, c):
            a, b, c = a, b, c
            a, c, b = a, b, c
            b, a, c = a, b, c
            b, c, a = a, b, c
            c, a, b = a, b, c
            c, b, a = a, b, c
        self.assertNotInBytecode(f, "SWAP")

    def test_static_swaps_match_mapping(self):
        for a, b, c in product("_a", "_b", "_c"):
            pattern = f"{{'a': {a}, 'b': {b}, 'c': {c}}}"
            with self.subTest(pattern):
                code = compile_pattern_with_fast_locals(pattern)
                self.assertNotInBytecode(code, "SWAP")

    def test_static_swaps_match_class(self):
        forms = [
            "C({}, {}, {})",
            "C({}, {}, c={})",
            "C({}, b={}, c={})",
            "C(a={}, b={}, c={})"
        ]
        for a, b, c in product("_a", "_b", "_c"):
            for form in forms:
                pattern = form.format(a, b, c)
                with self.subTest(pattern):
                    code = compile_pattern_with_fast_locals(pattern)
                    self.assertNotInBytecode(code, "SWAP")

    def test_static_swaps_match_sequence(self):
        swaps = {"*_, b, c", "a, *_, c", "a, b, *_"}
        forms = ["{}, {}, {}", "{}, {}, *{}", "{}, *{}, {}", "*{}, {}, {}"]
        for a, b, c in product("_a", "_b", "_c"):
            for form in forms:
                pattern = form.format(a, b, c)
                with self.subTest(pattern):
                    code = compile_pattern_with_fast_locals(pattern)
                    if pattern in swaps:
                        # If this fails... great! Remove this pattern from swaps
                        # to prevent regressing on any improvement:
                        self.assertInBytecode(code, "SWAP")
                    else:
                        self.assertNotInBytecode(code, "SWAP")


class TestBuglets(unittest.TestCase):

    def test_bug_11510(self):
        # folded constant set optimization was commingled with the tuple
        # unpacking optimization which would fail if the set had duplicate
        # elements so that the set length was unexpected
        def f():
            x, y = {1, 1}
            return x, y
        with self.assertRaises(ValueError):
            f()

    def test_bpo_42057(self):
        for i in range(10):
            try:
                raise Exception
            except Exception or Exception:
                pass

    def test_bpo_45773_pop_jump_if_true(self):
        compile("while True or spam: pass", "<test>", "exec")

    def test_bpo_45773_pop_jump_if_false(self):
        compile("while True or not spam: pass", "<test>", "exec")


class TestMarkingVariablesAsUnKnown(BytecodeTestCase):

    def setUp(self):
        self.addCleanup(sys.settrace, sys.gettrace())
        sys.settrace(None)

    def test_load_fast_known_simple(self):
        def f():
            x = 1
            y = x + x
        self.assertInBytecode(f, 'LOAD_FAST_BORROW_LOAD_FAST_BORROW')

    def test_load_fast_unknown_simple(self):
        def f():
            if condition():
                x = 1
            print(x)
        self.assertInBytecode(f, 'LOAD_FAST_CHECK')
        self.assertNotInBytecode(f, 'LOAD_FAST')

    def test_load_fast_unknown_because_del(self):
        def f():
            x = 1
            del x
            print(x)
        self.assertInBytecode(f, 'LOAD_FAST_CHECK')
        self.assertNotInBytecode(f, 'LOAD_FAST')

    def test_load_fast_known_because_parameter(self):
        def f1(x):
            print(x)
        self.assertInBytecode(f1, 'LOAD_FAST_BORROW')
        self.assertNotInBytecode(f1, 'LOAD_FAST_CHECK')

        def f2(*, x):
            print(x)
        self.assertInBytecode(f2, 'LOAD_FAST_BORROW')
        self.assertNotInBytecode(f2, 'LOAD_FAST_CHECK')

        def f3(*args):
            print(args)
        self.assertInBytecode(f3, 'LOAD_FAST_BORROW')
        self.assertNotInBytecode(f3, 'LOAD_FAST_CHECK')

        def f4(**kwargs):
            print(kwargs)
        self.assertInBytecode(f4, 'LOAD_FAST_BORROW')
        self.assertNotInBytecode(f4, 'LOAD_FAST_CHECK')

        def f5(x=0):
            print(x)
        self.assertInBytecode(f5, 'LOAD_FAST_BORROW')
        self.assertNotInBytecode(f5, 'LOAD_FAST_CHECK')

    def test_load_fast_known_because_already_loaded(self):
        def f():
            if condition():
                x = 1
            print(x)
            print(x)
        self.assertInBytecode(f, 'LOAD_FAST_CHECK')
        self.assertInBytecode(f, 'LOAD_FAST_BORROW')

    def test_load_fast_known_multiple_branches(self):
        def f():
            if condition():
                x = 1
            else:
                x = 2
            print(x)
        self.assertInBytecode(f, 'LOAD_FAST_BORROW')
        self.assertNotInBytecode(f, 'LOAD_FAST_CHECK')

    def test_load_fast_unknown_after_error(self):
        def f():
            try:
                res = 1 / 0
            except ZeroDivisionError:
                pass
            return res
        # LOAD_FAST (known) still occurs in the no-exception branch.
        # Assert that it doesn't occur in the LOAD_FAST_CHECK branch.
        self.assertInBytecode(f, 'LOAD_FAST_CHECK')

    def test_load_fast_unknown_after_error_2(self):
        def f():
            try:
                1 / 0
            except ZeroDivisionError:
                print(a, b, c, d, e, f, g)
            a = b = c = d = e = f = g = 1
        self.assertInBytecode(f, 'LOAD_FAST_CHECK')
        self.assertNotInBytecode(f, 'LOAD_FAST')

    def test_load_fast_too_many_locals(self):
        # When there get to be too many locals to analyze completely,
        # later locals are all converted to LOAD_FAST_CHECK, except
        # when a store or prior load occurred in the same basicblock.
        def f():
            a00 = a01 = a02 = a03 = a04 = a05 = a06 = a07 = a08 = a09 = 1
            a10 = a11 = a12 = a13 = a14 = a15 = a16 = a17 = a18 = a19 = 1
            a20 = a21 = a22 = a23 = a24 = a25 = a26 = a27 = a28 = a29 = 1
            a30 = a31 = a32 = a33 = a34 = a35 = a36 = a37 = a38 = a39 = 1
            a40 = a41 = a42 = a43 = a44 = a45 = a46 = a47 = a48 = a49 = 1
            a50 = a51 = a52 = a53 = a54 = a55 = a56 = a57 = a58 = a59 = 1
            a60 = a61 = a62 = a63 = a64 = a65 = a66 = a67 = a68 = a69 = 1
            a70 = a71 = a72 = a73 = a74 = a75 = a76 = a77 = a78 = a79 = 1
            del a72, a73
            print(a73)
            print(a70, a71, a72, a73)
            while True:
                print(a00, a01, a62, a63)
                print(a64, a65, a78, a79)

        self.assertInBytecode(f, 'LOAD_FAST_BORROW_LOAD_FAST_BORROW', ("a00", "a01"))
        self.assertNotInBytecode(f, 'LOAD_FAST_CHECK', "a00")
        self.assertNotInBytecode(f, 'LOAD_FAST_CHECK', "a01")
        for i in 62, 63:
            # First 64 locals: analyze completely
            self.assertInBytecode(f, 'LOAD_FAST_BORROW', f"a{i:02}")
            self.assertNotInBytecode(f, 'LOAD_FAST_CHECK', f"a{i:02}")
        for i in 64, 65, 78, 79:
            # Locals >=64 not in the same basicblock
            self.assertInBytecode(f, 'LOAD_FAST_CHECK', f"a{i:02}")
            self.assertNotInBytecode(f, 'LOAD_FAST', f"a{i:02}")
        for i in 70, 71:
            # Locals >=64 in the same basicblock
            self.assertInBytecode(f, 'LOAD_FAST_BORROW', f"a{i:02}")
            self.assertNotInBytecode(f, 'LOAD_FAST_CHECK', f"a{i:02}")
        # del statements should invalidate within basicblocks.
        self.assertInBytecode(f, 'LOAD_FAST_CHECK', "a72")
        self.assertNotInBytecode(f, 'LOAD_FAST', "a72")
        # previous checked loads within a basicblock enable unchecked loads
        self.assertInBytecode(f, 'LOAD_FAST_CHECK', "a73")
        self.assertInBytecode(f, 'LOAD_FAST_BORROW', "a73")

    def test_setting_lineno_no_undefined(self):
        code = textwrap.dedent("""\
            def f():
                x = y = 2
                if not x:
                    return 4
                for i in range(55):
                    x + 6
                L = 7
                L = 8
                L = 9
                L = 10
        """)
        ns = {}
        exec(code, ns)
        f = ns['f']
        self.assertInBytecode(f, "LOAD_FAST_BORROW")
        self.assertNotInBytecode(f, "LOAD_FAST_CHECK")
        co_code = f.__code__.co_code
        def trace(frame, event, arg):
            if event == 'line' and frame.f_lineno == 9:
                frame.f_lineno = 3
                sys.settrace(None)
                return None
            return trace
        sys.settrace(trace)
        result = f()
        self.assertIsNone(result)
        self.assertInBytecode(f, "LOAD_FAST_BORROW")
        self.assertNotInBytecode(f, "LOAD_FAST_CHECK")
        self.assertEqual(f.__code__.co_code, co_code)

    def test_setting_lineno_one_undefined(self):
        code = textwrap.dedent("""\
            def f():
                x = y = 2
                if not x:
                    return 4
                for i in range(55):
                    x + 6
                del x
                L = 8
                L = 9
                L = 10
        """)
        ns = {}
        exec(code, ns)
        f = ns['f']
        self.assertInBytecode(f, "LOAD_FAST_BORROW")
        self.assertNotInBytecode(f, "LOAD_FAST_CHECK")
        co_code = f.__code__.co_code
        def trace(frame, event, arg):
            if event == 'line' and frame.f_lineno == 9:
                frame.f_lineno = 3
                sys.settrace(None)
                return None
            return trace
        e = r"assigning None to 1 unbound local"
        with self.assertWarnsRegex(RuntimeWarning, e):
            sys.settrace(trace)
            result = f()
        self.assertEqual(result, 4)
        self.assertInBytecode(f, "LOAD_FAST_BORROW")
        self.assertNotInBytecode(f, "LOAD_FAST_CHECK")
        self.assertEqual(f.__code__.co_code, co_code)

    def test_setting_lineno_two_undefined(self):
        code = textwrap.dedent("""\
            def f():
                x = y = 2
                if not x:
                    return 4
                for i in range(55):
                    x + 6
                del x, y
                L = 8
                L = 9
                L = 10
        """)
        ns = {}
        exec(code, ns)
        f = ns['f']
        self.assertInBytecode(f, "LOAD_FAST_BORROW")
        self.assertNotInBytecode(f, "LOAD_FAST_CHECK")
        co_code = f.__code__.co_code
        def trace(frame, event, arg):
            if event == 'line' and frame.f_lineno == 9:
                frame.f_lineno = 3
                sys.settrace(None)
                return None
            return trace
        e = r"assigning None to 2 unbound locals"
        with self.assertWarnsRegex(RuntimeWarning, e):
            sys.settrace(trace)
            result = f()
        self.assertEqual(result, 4)
        self.assertInBytecode(f, "LOAD_FAST_BORROW")
        self.assertNotInBytecode(f, "LOAD_FAST_CHECK")
        self.assertEqual(f.__code__.co_code, co_code)

    def make_function_with_no_checks(self):
        code = textwrap.dedent("""\
            def f():
                x = 2
                L = 3
                L = 4
                L = 5
                if not L:
                    x + 7
                    y = 2
        """)
        ns = {}
        exec(code, ns)
        f = ns['f']
        self.assertInBytecode(f, "LOAD_FAST_BORROW")
        self.assertNotInBytecode(f, "LOAD_FAST_CHECK")
        return f

    def test_modifying_local_does_not_add_check(self):
        f = self.make_function_with_no_checks()
        def trace(frame, event, arg):
            if event == 'line' and frame.f_lineno == 4:
                frame.f_locals["x"] = 42
                sys.settrace(None)
                return None
            return trace
        sys.settrace(trace)
        f()
        self.assertInBytecode(f, "LOAD_FAST_BORROW")
        self.assertNotInBytecode(f, "LOAD_FAST_CHECK")

    def test_initializing_local_does_not_add_check(self):
        f = self.make_function_with_no_checks()
        def trace(frame, event, arg):
            if event == 'line' and frame.f_lineno == 4:
                frame.f_locals["y"] = 42
                sys.settrace(None)
                return None
            return trace
        sys.settrace(trace)
        f()
        self.assertInBytecode(f, "LOAD_FAST_BORROW")
        self.assertNotInBytecode(f, "LOAD_FAST_CHECK")


class DirectCfgOptimizerTests(CfgOptimizationTestCase):

    def cfg_optimization_test(self, insts, expected_insts,
                              consts=None, expected_consts=None,
                              nlocals=0):

        self.check_instructions(insts)
        self.check_instructions(expected_insts)

        if expected_consts is None:
            expected_consts = consts
        seq = self.seq_from_insts(insts)
        opt_insts, opt_consts = self.get_optimized(seq, consts, nlocals)
        expected_insts = self.seq_from_insts(expected_insts).get_instructions()
        self.assertInstructionsMatch(opt_insts, expected_insts)
        self.assertEqual(opt_consts, expected_consts)

    def test_conditional_jump_forward_non_const_condition(self):
        insts = [
            ('LOAD_NAME', 1, 11),
            ('POP_JUMP_IF_TRUE', lbl := self.Label(), 12),
            ('LOAD_CONST', 2, 13),
            ('RETURN_VALUE', None, 13),
            lbl,
            ('LOAD_CONST', 3, 14),
            ('RETURN_VALUE', None, 14),
        ]
        expected_insts = [
            ('LOAD_NAME', 1, 11),
            ('POP_JUMP_IF_TRUE', lbl := self.Label(), 12),
            ('LOAD_SMALL_INT', 2, 13),
            ('RETURN_VALUE', None, 13),
            lbl,
            ('LOAD_SMALL_INT', 3, 14),
            ('RETURN_VALUE', None, 14),
        ]
        self.cfg_optimization_test(insts,
                                   expected_insts,
                                   consts=[0, 1, 2, 3, 4],
                                   expected_consts=[0])

    def test_list_exceeding_stack_use_guideline(self):
        def f():
            return [
                0, 1, 2, 3, 4,
                5, 6, 7, 8, 9,
                10, 11, 12, 13, 14,
                15, 16, 17, 18, 19,
                20, 21, 22, 23, 24,
                25, 26, 27, 28, 29,
                30, 31, 32, 33, 34,
                35, 36, 37, 38, 39
            ]
        self.assertEqual(f(), list(range(40)))

    def test_set_exceeding_stack_use_guideline(self):
        def f():
            return {
                0, 1, 2, 3, 4,
                5, 6, 7, 8, 9,
                10, 11, 12, 13, 14,
                15, 16, 17, 18, 19,
                20, 21, 22, 23, 24,
                25, 26, 27, 28, 29,
                30, 31, 32, 33, 34,
                35, 36, 37, 38, 39
            }
        self.assertEqual(f(), frozenset(range(40)))

    def test_nested_const_foldings(self):
        # (1, (--2 + ++2 * 2 // 2 - 2, )[0], ~~3, not not True)  ==>  (1, 2, 3, True)
        intrinsic_positive = 5
        before = [
            ('LOAD_SMALL_INT', 1, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('UNARY_NEGATIVE', None, 0),
            ('NOP', None, 0),
            ('UNARY_NEGATIVE', None, 0),
            ('NOP', None, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('CALL_INTRINSIC_1', intrinsic_positive, 0),
            ('NOP', None, 0),
            ('CALL_INTRINSIC_1', intrinsic_positive, 0),
            ('BINARY_OP', get_binop_argval('NB_MULTIPLY')),
            ('LOAD_SMALL_INT', 2, 0),
            ('NOP', None, 0),
            ('BINARY_OP', get_binop_argval('NB_FLOOR_DIVIDE')),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', get_binop_argval('NB_ADD')),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('NOP', None, 0),
            ('BINARY_OP', get_binop_argval('NB_SUBTRACT')),
            ('NOP', None, 0),
            ('BUILD_TUPLE', 1, 0),
            ('LOAD_SMALL_INT', 0, 0),
            ('BINARY_OP', get_binop_argval('NB_SUBSCR'), 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 3, 0),
            ('NOP', None, 0),
            ('UNARY_INVERT', None, 0),
            ('NOP', None, 0),
            ('UNARY_INVERT', None, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 3, 0),
            ('NOP', None, 0),
            ('UNARY_NOT', None, 0),
            ('NOP', None, 0),
            ('UNARY_NOT', None, 0),
            ('NOP', None, 0),
            ('BUILD_TUPLE', 4, 0),
            ('NOP', None, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_CONST', 1, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[-2, (1, 2, 3, True)])

    def test_build_empty_tuple(self):
        before = [
            ('BUILD_TUPLE', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[()])

    def test_fold_tuple_of_constants(self):
        before = [
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('NOP', None, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 3, 0),
            ('NOP', None, 0),
            ('BUILD_TUPLE', 3, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[(1, 2, 3)])

        # not all consts
        same = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_NAME', 0, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BUILD_TUPLE', 3, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(same, same, consts=[])

    def test_fold_constant_intrinsic_list_to_tuple(self):
        INTRINSIC_LIST_TO_TUPLE = 6

        # long tuple
        consts = 1000
        before = (
            [('BUILD_LIST', 0, 0)] +
            [('LOAD_CONST', 0, 0), ('LIST_APPEND', 1, 0)] * consts +
            [('CALL_INTRINSIC_1', INTRINSIC_LIST_TO_TUPLE, 0), ('RETURN_VALUE', None, 0)]
        )
        after = [
            ('LOAD_CONST', 1, 0),
            ('RETURN_VALUE', None, 0)
        ]
        result_const = tuple(["test"] * consts)
        self.cfg_optimization_test(before, after, consts=["test"], expected_consts=["test", result_const])

        # empty list
        before = [
            ('BUILD_LIST', 0, 0),
            ('CALL_INTRINSIC_1', INTRINSIC_LIST_TO_TUPLE, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[()])

        # multiple BUILD_LIST 0: ([], 1, [], 2)
        same = [
            ('BUILD_LIST', 0, 0),
            ('BUILD_LIST', 0, 0),
            ('LIST_APPEND', 1, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('LIST_APPEND', 1, 0),
            ('BUILD_LIST', 0, 0),
            ('LIST_APPEND', 1, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('LIST_APPEND', 1, 0),
            ('CALL_INTRINSIC_1', INTRINSIC_LIST_TO_TUPLE, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(same, same, consts=[])

        # nested folding: (1, 1+1, 3)
        before = [
            ('BUILD_LIST', 0, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('LIST_APPEND', 1, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('BINARY_OP', get_binop_argval('NB_ADD'), 0),
            ('LIST_APPEND', 1, 0),
            ('LOAD_SMALL_INT', 3, 0),
            ('LIST_APPEND', 1, 0),
            ('CALL_INTRINSIC_1', INTRINSIC_LIST_TO_TUPLE, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[(1, 2, 3)])

        # NOP's in between: (1, 2, 3)
        before = [
            ('BUILD_LIST', 0, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('NOP', None, 0),
            ('NOP', None, 0),
            ('LIST_APPEND', 1, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('NOP', None, 0),
            ('NOP', None, 0),
            ('LIST_APPEND', 1, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 3, 0),
            ('NOP', None, 0),
            ('LIST_APPEND', 1, 0),
            ('NOP', None, 0),
            ('CALL_INTRINSIC_1', INTRINSIC_LIST_TO_TUPLE, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[(1, 2, 3)])

    def test_optimize_if_const_list(self):
        before = [
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('NOP', None, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 3, 0),
            ('NOP', None, 0),
            ('BUILD_LIST', 3, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('BUILD_LIST', 0, 0),
            ('LOAD_CONST', 0, 0),
            ('LIST_EXTEND', 1, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[(1, 2, 3)])

        # need minimum 3 consts to optimize
        same = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BUILD_LIST', 2, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(same, same, consts=[])

        # not all consts
        same = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_NAME', 0, 0),
            ('LOAD_SMALL_INT', 3, 0),
            ('BUILD_LIST', 3, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(same, same, consts=[])

    def test_optimize_if_const_set(self):
        before = [
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('NOP', None, 0),
            ('NOP', None, 0),
            ('LOAD_SMALL_INT', 3, 0),
            ('NOP', None, 0),
            ('BUILD_SET', 3, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('BUILD_SET', 0, 0),
            ('LOAD_CONST', 0, 0),
            ('SET_UPDATE', 1, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[frozenset({1, 2, 3})])

        # need minimum 3 consts to optimize
        same = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BUILD_SET', 2, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(same, same, consts=[])

        # not all consts
        same = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_NAME', 0, 0),
            ('LOAD_SMALL_INT', 3, 0),
            ('BUILD_SET', 3, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(same, same, consts=[])

    def test_optimize_literal_list_for_iter(self):
        # for _ in [1, 2]: pass  ==>  for _ in (1, 2): pass
        before = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BUILD_LIST', 2, 0),
            ('GET_ITER', None, 0),
            start := self.Label(),
            ('FOR_ITER', end := self.Label(), 0),
            ('STORE_FAST', 0, 0),
            ('JUMP', start, 0),
            end,
            ('END_FOR', None, 0),
            ('POP_ITER', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_CONST', 1, 0),
            ('GET_ITER', None, 0),
            start := self.Label(),
            ('FOR_ITER', end := self.Label(), 0),
            ('STORE_FAST', 0, 0),
            ('JUMP', start, 0),
            end,
            ('END_FOR', None, 0),
            ('POP_ITER', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[None], expected_consts=[None, (1, 2)])

        # for _ in [1, x]: pass  ==>  for _ in (1, x): pass
        before = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_NAME', 0, 0),
            ('BUILD_LIST', 2, 0),
            ('GET_ITER', None, 0),
            start := self.Label(),
            ('FOR_ITER', end := self.Label(), 0),
            ('STORE_FAST', 0, 0),
            ('JUMP', start, 0),
            end,
            ('END_FOR', None, 0),
            ('POP_ITER', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_NAME', 0, 0),
            ('BUILD_TUPLE', 2, 0),
            ('GET_ITER', None, 0),
            start := self.Label(),
            ('FOR_ITER', end := self.Label(), 0),
            ('STORE_FAST', 0, 0),
            ('JUMP', start, 0),
            end,
            ('END_FOR', None, 0),
            ('POP_ITER', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[None], expected_consts=[None])

    def test_optimize_literal_set_for_iter(self):
        # for _ in {1, 2}: pass  ==>  for _ in (1, 2): pass
        before = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BUILD_SET', 2, 0),
            ('GET_ITER', None, 0),
            start := self.Label(),
            ('FOR_ITER', end := self.Label(), 0),
            ('STORE_FAST', 0, 0),
            ('JUMP', start, 0),
            end,
            ('END_FOR', None, 0),
            ('POP_ITER', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_CONST', 1, 0),
            ('GET_ITER', None, 0),
            start := self.Label(),
            ('FOR_ITER', end := self.Label(), 0),
            ('STORE_FAST', 0, 0),
            ('JUMP', start, 0),
            end,
            ('END_FOR', None, 0),
            ('POP_ITER', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[None], expected_consts=[None, frozenset({1, 2})])

        # non constant literal set is not changed
        # for _ in {1, x}: pass  ==>  for _ in {1, x}: pass
        same = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_NAME', 0, 0),
            ('BUILD_SET', 2, 0),
            ('GET_ITER', None, 0),
            start := self.Label(),
            ('FOR_ITER', end := self.Label(), 0),
            ('STORE_FAST', 0, 0),
            ('JUMP', start, 0),
            end,
            ('END_FOR', None, 0),
            ('POP_ITER', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(same, same, consts=[None], expected_consts=[None])

    def test_optimize_literal_list_contains(self):
        # x in [1, 2]  ==>  x in (1, 2)
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BUILD_LIST', 2, 0),
            ('CONTAINS_OP', 0, 0),
            ('POP_TOP', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_CONST', 1, 0),
            ('CONTAINS_OP', 0, 0),
            ('POP_TOP', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[None], expected_consts=[None, (1, 2)])

        # x in [1, y]  ==>  x in (1, y)
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_NAME', 1, 0),
            ('BUILD_LIST', 2, 0),
            ('CONTAINS_OP', 0, 0),
            ('POP_TOP', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_NAME', 1, 0),
            ('BUILD_TUPLE', 2, 0),
            ('CONTAINS_OP', 0, 0),
            ('POP_TOP', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[None], expected_consts=[None])

    def test_optimize_literal_set_contains(self):
        # x in {1, 2}  ==>  x in (1, 2)
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BUILD_SET', 2, 0),
            ('CONTAINS_OP', 0, 0),
            ('POP_TOP', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_CONST', 1, 0),
            ('CONTAINS_OP', 0, 0),
            ('POP_TOP', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[None], expected_consts=[None, frozenset({1, 2})])

        # non constant literal set is not changed
        # x in {1, y}  ==>  x in {1, y}
        same = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_NAME', 1, 0),
            ('BUILD_SET', 2, 0),
            ('CONTAINS_OP', 0, 0),
            ('POP_TOP', None, 0),
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(same, same, consts=[None], expected_consts=[None])

    def test_optimize_unary_not(self):
        # test folding
        before = [
            ('LOAD_SMALL_INT', 1, 0),
            ('TO_BOOL', None, 0),
            ('UNARY_NOT', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_CONST', 1, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[True, False])

        # test cancel out
        before = [
            ('LOAD_NAME', 0, 0),
            ('TO_BOOL', None, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('TO_BOOL', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test eliminate to bool
        before = [
            ('LOAD_NAME', 0, 0),
            ('TO_BOOL', None, 0),
            ('UNARY_NOT', None, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('TO_BOOL', None, 0),
            ('UNARY_NOT', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test folding & cancel out
        before = [
            ('LOAD_SMALL_INT', 1, 0),
            ('TO_BOOL', None, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[True])

        # test folding & eliminate to bool
        before = [
            ('LOAD_SMALL_INT', 1, 0),
            ('TO_BOOL', None, 0),
            ('UNARY_NOT', None, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_CONST', 1, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[True, False])

        # test cancel out & eliminate to bool (to bool stays as we are not iterating to a fixed point)
        before = [
            ('LOAD_NAME', 0, 0),
            ('TO_BOOL', None, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('TO_BOOL', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        is_ = in_ = 0
        isnot = notin = 1

        # test is/isnot
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('IS_OP', is_, 0),
            ('UNARY_NOT', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('IS_OP', isnot, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test is/isnot cancel out
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('IS_OP', is_, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('IS_OP', is_, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test is/isnot eliminate to bool
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('IS_OP', is_, 0),
            ('UNARY_NOT', None, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('IS_OP', isnot, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test is/isnot cancel out & eliminate to bool
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('IS_OP', is_, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('IS_OP', is_, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test in/notin
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('CONTAINS_OP', in_, 0),
            ('UNARY_NOT', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('CONTAINS_OP', notin, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test in/notin cancel out
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('CONTAINS_OP', in_, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('CONTAINS_OP', in_, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test is/isnot & eliminate to bool
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('CONTAINS_OP', in_, 0),
            ('UNARY_NOT', None, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('TO_BOOL', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('CONTAINS_OP', notin, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test in/notin cancel out & eliminate to bool
        before = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('CONTAINS_OP', in_, 0),
            ('UNARY_NOT', None, 0),
            ('UNARY_NOT', None, 0),
            ('TO_BOOL', None, 0),
            ('RETURN_VALUE', None, 0),
        ]
        after = [
            ('LOAD_NAME', 0, 0),
            ('LOAD_NAME', 1, 0),
            ('CONTAINS_OP', in_, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

    def test_optimize_if_const_unaryop(self):
        # test unary negative
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('UNARY_NEGATIVE', None, 0),
            ('UNARY_NEGATIVE', None, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 2, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[-2])

        # test unary invert
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('UNARY_INVERT', None, 0),
            ('UNARY_INVERT', None, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 2, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[-3])

        # test unary positive
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('CALL_INTRINSIC_1', 5, 0),
            ('CALL_INTRINSIC_1', 5, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 2, 0),
            ('RETURN_VALUE', None, 0),
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

    def test_optimize_if_const_binop(self):
        add = get_binop_argval('NB_ADD')
        sub = get_binop_argval('NB_SUBTRACT')
        mul = get_binop_argval('NB_MULTIPLY')
        div = get_binop_argval('NB_TRUE_DIVIDE')
        floor = get_binop_argval('NB_FLOOR_DIVIDE')
        rem = get_binop_argval('NB_REMAINDER')
        pow = get_binop_argval('NB_POWER')
        lshift = get_binop_argval('NB_LSHIFT')
        rshift = get_binop_argval('NB_RSHIFT')
        or_ = get_binop_argval('NB_OR')
        and_ = get_binop_argval('NB_AND')
        xor = get_binop_argval('NB_XOR')
        subscr = get_binop_argval('NB_SUBSCR')

        # test add
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', add, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', add, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 6, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test sub
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', sub, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', sub, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_CONST', 0, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[-2])

        # test mul
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', mul, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', mul, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 8, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test div
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', div, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', div, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_CONST', 1, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[1.0, 0.5])

        # test floor
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', floor, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', floor, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 0, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test rem
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', rem, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', rem, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 0, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test pow
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', pow, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', pow, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 16, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test lshift
        before = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('BINARY_OP', lshift, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('BINARY_OP', lshift, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 4, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test rshift
        before = [
            ('LOAD_SMALL_INT', 4, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('BINARY_OP', rshift, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('BINARY_OP', rshift, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 1, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test or
        before = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', or_, 0),
            ('LOAD_SMALL_INT', 4, 0),
            ('BINARY_OP', or_, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 7, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test and
        before = [
            ('LOAD_SMALL_INT', 1, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('BINARY_OP', and_, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('BINARY_OP', and_, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 1, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test xor
        before = [
            ('LOAD_SMALL_INT', 2, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', xor, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', xor, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 2, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[], expected_consts=[])

        # test subscr
        before = [
            ('LOAD_CONST', 0, 0),
            ('LOAD_SMALL_INT', 1, 0),
            ('BINARY_OP', subscr, 0),
            ('LOAD_SMALL_INT', 2, 0),
            ('BINARY_OP', subscr, 0),
            ('RETURN_VALUE', None, 0)
        ]
        after = [
            ('LOAD_SMALL_INT', 3, 0),
            ('RETURN_VALUE', None, 0)
        ]
        self.cfg_optimization_test(before, after, consts=[(1, (1, 2, 3))], expected_consts=[(1, (1, 2, 3))])


    def test_conditional_jump_forward_const_condition(self):
        # The unreachable branch of the jump is removed, the jump
        # becomes redundant and is replaced by a NOP (for the lineno)

        insts = [
            ('LOAD_CONST', 3, 11),
            ('POP_JUMP_IF_TRUE', lbl := self.Label(), 12),
            ('LOAD_CONST', 2, 13),
            lbl,
            ('LOAD_CONST', 3, 14),
            ('RETURN_VALUE', None, 14),
        ]
        expected_insts = [
            ('NOP', None, 11),
            ('NOP', None, 12),
            ('LOAD_SMALL_INT', 3, 14),
            ('RETURN_VALUE', None, 14),
        ]
        self.cfg_optimization_test(insts,
                                   expected_insts,
                                   consts=[0, 1, 2, 3, 4],
                                   expected_consts=[0])

    def test_conditional_jump_backward_non_const_condition(self):
        insts = [
            lbl1 := self.Label(),
            ('LOAD_NAME', 1, 11),
            ('POP_JUMP_IF_TRUE', lbl1, 12),
            ('LOAD_NAME', 2, 13),
            ('RETURN_VALUE', None, 13),
        ]
        expected = [
            lbl := self.Label(),
            ('LOAD_NAME', 1, 11),
            ('POP_JUMP_IF_TRUE', lbl, 12),
            ('LOAD_NAME', 2, 13),
            ('RETURN_VALUE', None, 13),
        ]
        self.cfg_optimization_test(insts, expected, consts=list(range(5)))

    def test_conditional_jump_backward_const_condition(self):
        # The unreachable branch of the jump is removed
        insts = [
            lbl1 := self.Label(),
            ('LOAD_CONST', 3, 11),
            ('POP_JUMP_IF_TRUE', lbl1, 12),
            ('LOAD_CONST', 2, 13),
            ('RETURN_VALUE', None, 13),
        ]
        expected_insts = [
            lbl := self.Label(),
            ('NOP', None, 11),
            ('JUMP', lbl, 12),
        ]
        self.cfg_optimization_test(insts, expected_insts, consts=list(range(5)))

    def test_except_handler_label(self):
        insts = [
            ('SETUP_FINALLY', handler := self.Label(), 10),
            ('POP_BLOCK', None, -1),
            ('LOAD_CONST', 1, 11),
            ('RETURN_VALUE', None, 11),
            handler,
            ('LOAD_CONST', 2, 12),
            ('RETURN_VALUE', None, 12),
        ]
        expected_insts = [
            ('SETUP_FINALLY', handler := self.Label(), 10),
            ('LOAD_SMALL_INT', 1, 11),
            ('RETURN_VALUE', None, 11),
            handler,
            ('LOAD_SMALL_INT', 2, 12),
            ('RETURN_VALUE', None, 12),
        ]
        self.cfg_optimization_test(insts, expected_insts, consts=list(range(5)))

    def test_no_unsafe_static_swap(self):
        # We can't change order of two stores to the same location
        insts = [
            ('LOAD_CONST', 0, 1),
            ('LOAD_CONST', 1, 2),
            ('LOAD_CONST', 2, 3),
            ('SWAP', 3, 4),
            ('STORE_FAST', 1, 4),
            ('STORE_FAST', 1, 4),
            ('POP_TOP', None, 4),
            ('LOAD_CONST', 0, 5),
            ('RETURN_VALUE', None, 5)
        ]
        expected_insts = [
            ('LOAD_SMALL_INT', 0, 1),
            ('LOAD_SMALL_INT', 1, 2),
            ('NOP', None, 3),
            ('STORE_FAST', 1, 4),
            ('POP_TOP', None, 4),
            ('LOAD_SMALL_INT', 0, 5),
            ('RETURN_VALUE', None, 5)
        ]
        self.cfg_optimization_test(insts, expected_insts, consts=list(range(3)), nlocals=1)

    def test_dead_store_elimination_in_same_lineno(self):
        insts = [
            ('LOAD_CONST', 0, 1),
            ('LOAD_CONST', 1, 2),
            ('LOAD_CONST', 2, 3),
            ('STORE_FAST', 1, 4),
            ('STORE_FAST', 1, 4),
            ('STORE_FAST', 1, 4),
            ('LOAD_CONST', 0, 5),
            ('RETURN_VALUE', None, 5)
        ]
        expected_insts = [
            ('LOAD_SMALL_INT', 0, 1),
            ('LOAD_SMALL_INT', 1, 2),
            ('NOP', None, 3),
            ('POP_TOP', None, 4),
            ('STORE_FAST', 1, 4),
            ('LOAD_SMALL_INT', 0, 5),
            ('RETURN_VALUE', None, 5)
        ]
        self.cfg_optimization_test(insts, expected_insts, consts=list(range(3)), nlocals=1)

    def test_no_dead_store_elimination_in_different_lineno(self):
        insts = [
            ('LOAD_CONST', 0, 1),
            ('LOAD_CONST', 1, 2),
            ('LOAD_CONST', 2, 3),
            ('STORE_FAST', 1, 4),
            ('STORE_FAST', 1, 5),
            ('STORE_FAST', 1, 6),
            ('LOAD_CONST', 0, 5),
            ('RETURN_VALUE', None, 5)
        ]
        expected_insts = [
            ('LOAD_SMALL_INT', 0, 1),
            ('LOAD_SMALL_INT', 1, 2),
            ('LOAD_SMALL_INT', 2, 3),
            ('STORE_FAST', 1, 4),
            ('STORE_FAST', 1, 5),
            ('STORE_FAST', 1, 6),
            ('LOAD_SMALL_INT', 0, 5),
            ('RETURN_VALUE', None, 5)
        ]
        self.cfg_optimization_test(insts, expected_insts, consts=list(range(3)), nlocals=1)

    def test_unconditional_jump_threading(self):

        def get_insts(lno1, lno2, op1, op2):
            return [
                       lbl2 := self.Label(),
                       ('LOAD_NAME', 0, 10),
                       ('POP_TOP', None, 10),
                       (op1, lbl1 := self.Label(), lno1),
                       ('LOAD_NAME', 1, 20),
                       lbl1,
                       (op2, lbl2, lno2),
                   ]

        for op1 in ('JUMP', 'JUMP_NO_INTERRUPT'):
            for op2 in ('JUMP', 'JUMP_NO_INTERRUPT'):
                # different lines
                lno1, lno2 = (4, 5)
                with self.subTest(lno = (lno1, lno2), ops = (op1, op2)):
                    insts = get_insts(lno1, lno2, op1, op2)
                    op = 'JUMP' if 'JUMP' in (op1, op2) else 'JUMP_NO_INTERRUPT'
                    expected_insts = [
                        ('LOAD_NAME', 0, 10),
                        ('POP_TOP', None, 10),
                        ('NOP', None, 4),
                        (op, 0, 5),
                    ]
                    self.cfg_optimization_test(insts, expected_insts, consts=list(range(5)))

                # Threading
                for lno1, lno2 in [(-1, -1), (-1, 5), (6, -1), (7, 7)]:
                    with self.subTest(lno = (lno1, lno2), ops = (op1, op2)):
                        insts = get_insts(lno1, lno2, op1, op2)
                        lno = lno1 if lno1 != -1 else lno2
                        if lno == -1:
                            lno = 10  # Propagated from the line before

                        op = 'JUMP' if 'JUMP' in (op1, op2) else 'JUMP_NO_INTERRUPT'
                        expected_insts = [
                            ('LOAD_NAME', 0, 10),
                            ('POP_TOP', None, 10),
                            (op, 0, lno),
                        ]
                        self.cfg_optimization_test(insts, expected_insts, consts=list(range(5)))

    def test_list_to_tuple_get_iter(self):
        # for _ in (*foo, *bar) -> for _ in [*foo, *bar]
        INTRINSIC_LIST_TO_TUPLE = 6
        insts = [
            ("BUILD_LIST", 0, 1),
            ("LOAD_FAST", 0, 2),
            ("LIST_EXTEND", 1, 3),
            ("LOAD_FAST", 1, 4),
            ("LIST_EXTEND", 1, 5),
            ("CALL_INTRINSIC_1", INTRINSIC_LIST_TO_TUPLE, 6),
            ("GET_ITER", None, 7),
            top := self.Label(),
            ("FOR_ITER", end := self.Label(), 8),
            ("STORE_FAST", 2, 9),
            ("JUMP", top, 10),
            end,
            ("END_FOR", None, 11),
            ("POP_TOP", None, 12),
            ("LOAD_CONST", 0, 13),
            ("RETURN_VALUE", None, 14),
        ]
        expected_insts = [
            ("BUILD_LIST", 0, 1),
            ("LOAD_FAST_BORROW", 0, 2),
            ("LIST_EXTEND", 1, 3),
            ("LOAD_FAST_BORROW", 1, 4),
            ("LIST_EXTEND", 1, 5),
            ("NOP", None, 6),  # ("CALL_INTRINSIC_1", INTRINSIC_LIST_TO_TUPLE, 6),
            ("GET_ITER", None, 7),
            top := self.Label(),
            ("FOR_ITER", end := self.Label(), 8),
            ("STORE_FAST", 2, 9),
            ("JUMP", top, 10),
            end,
            ("END_FOR", None, 11),
            ("POP_TOP", None, 12),
            ("LOAD_CONST", 0, 13),
            ("RETURN_VALUE", None, 14),
        ]
        self.cfg_optimization_test(insts, expected_insts, consts=[None])

    def test_list_to_tuple_get_iter_is_safe(self):
        a, b = [], []
        for item in (*(items := [0, 1, 2, 3]),):
            a.append(item)
            b.append(items.pop())
        self.assertEqual(a, [0, 1, 2, 3])
        self.assertEqual(b, [3, 2, 1, 0])
        self.assertEqual(items, [])


class OptimizeLoadFastTestCase(DirectCfgOptimizerTests):
    def make_bb(self, insts):
        last_loc = insts[-1][2]
        maxconst = 0
        for op, arg, _ in insts:
            if op == "LOAD_CONST":
                maxconst = max(maxconst, arg)
        consts = [None for _ in range(maxconst + 1)]
        return insts + [
            ("LOAD_CONST", 0, last_loc + 1),
            ("RETURN_VALUE", None, last_loc + 2),
        ], consts

    def check(self, insts, expected_insts, consts=None):
        insts_bb, insts_consts = self.make_bb(insts)
        expected_insts_bb, exp_consts = self.make_bb(expected_insts)
        self.cfg_optimization_test(insts_bb, expected_insts_bb,
                                   consts=insts_consts, expected_consts=exp_consts)

    def test_optimized(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_FAST", 1, 2),
            ("BINARY_OP", 2, 3),
        ]
        expected = [
            ("LOAD_FAST_BORROW", 0, 1),
            ("LOAD_FAST_BORROW", 1, 2),
            ("BINARY_OP", 2, 3),
        ]
        self.check(insts, expected)

        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_CONST", 1, 2),
            ("SWAP", 2, 3),
            ("POP_TOP", None, 4),
        ]
        expected = [
            ("LOAD_FAST_BORROW", 0, 1),
            ("LOAD_CONST", 1, 2),
            ("SWAP", 2, 3),
            ("POP_TOP", None, 4),
        ]
        self.check(insts, expected)

    def test_unoptimized_if_unconsumed(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_FAST", 1, 2),
            ("POP_TOP", None, 3),
        ]
        expected = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_FAST_BORROW", 1, 2),
            ("POP_TOP", None, 3),
        ]
        self.check(insts, expected)

        insts = [
            ("LOAD_FAST", 0, 1),
            ("COPY", 1, 2),
            ("POP_TOP", None, 3),
        ]
        expected = [
            ("LOAD_FAST", 0, 1),
            ("NOP", None, 2),
            ("NOP", None, 3),
        ]
        self.check(insts, expected)

    def test_unoptimized_if_support_killed(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_CONST", 0, 2),
            ("STORE_FAST", 0, 3),
            ("POP_TOP", None, 4),
        ]
        self.check(insts, insts)

        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_CONST", 0, 2),
            ("LOAD_CONST", 0, 3),
            ("STORE_FAST_STORE_FAST", ((0 << 4) | 1), 4),
            ("POP_TOP", None, 5),
        ]
        self.check(insts, insts)

        insts = [
            ("LOAD_FAST", 0, 1),
            ("DELETE_FAST", 0, 2),
            ("POP_TOP", None, 3),
        ]
        self.check(insts, insts)

    def test_unoptimized_if_aliased(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("STORE_FAST", 1, 2),
        ]
        self.check(insts, insts)

        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_CONST", 0, 3),
            ("STORE_FAST_STORE_FAST", ((0 << 4) | 1), 4),
        ]
        self.check(insts, insts)

    def test_consume_no_inputs(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("GET_LEN", None, 2),
            ("STORE_FAST", 1 , 3),
            ("STORE_FAST", 2, 4),
        ]
        self.check(insts, insts)

    def test_consume_some_inputs_no_outputs(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("GET_LEN", None, 2),
            ("LIST_APPEND", 0, 3),
        ]
        self.check(insts, insts)

    def test_check_exc_match(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_FAST", 1, 2),
            ("CHECK_EXC_MATCH", None, 3)
        ]
        expected = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_FAST_BORROW", 1, 2),
            ("CHECK_EXC_MATCH", None, 3)
        ]
        self.check(insts, expected)

    def test_for_iter(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            top := self.Label(),
            ("FOR_ITER", end := self.Label(), 2),
            ("STORE_FAST", 2, 3),
            ("JUMP", top, 4),
            end,
            ("END_FOR", None, 5),
            ("POP_TOP", None, 6),
            ("LOAD_CONST", 0, 7),
            ("RETURN_VALUE", None, 8),
        ]
        self.cfg_optimization_test(insts, insts, consts=[None])

    def test_load_attr(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_ATTR", 0, 2),
        ]
        expected = [
            ("LOAD_FAST_BORROW", 0, 1),
            ("LOAD_ATTR", 0, 2),
        ]
        self.check(insts, expected)

        # Method call, leaves self on stack unconsumed
        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_ATTR", 1, 2),
        ]
        expected = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_ATTR", 1, 2),
        ]
        self.check(insts, expected)

    def test_super_attr(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_FAST", 1, 2),
            ("LOAD_FAST", 2, 3),
            ("LOAD_SUPER_ATTR", 0, 4),
        ]
        expected = [
            ("LOAD_FAST_BORROW", 0, 1),
            ("LOAD_FAST_BORROW", 1, 2),
            ("LOAD_FAST_BORROW", 2, 3),
            ("LOAD_SUPER_ATTR", 0, 4),
        ]
        self.check(insts, expected)

        # Method call, leaves self on stack unconsumed
        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_FAST", 1, 2),
            ("LOAD_FAST", 2, 3),
            ("LOAD_SUPER_ATTR", 1, 4),
        ]
        expected = [
            ("LOAD_FAST_BORROW", 0, 1),
            ("LOAD_FAST_BORROW", 1, 2),
            ("LOAD_FAST", 2, 3),
            ("LOAD_SUPER_ATTR", 1, 4),
        ]
        self.check(insts, expected)

    def test_send(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_FAST", 1, 2),
            ("SEND", end := self.Label(), 3),
            ("LOAD_CONST", 0, 4),
            ("RETURN_VALUE", None, 5),
            end,
            ("LOAD_CONST", 0, 6),
            ("RETURN_VALUE", None, 7)
        ]
        expected = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_FAST_BORROW", 1, 2),
            ("SEND", end := self.Label(), 3),
            ("LOAD_CONST", 0, 4),
            ("RETURN_VALUE", None, 5),
            end,
            ("LOAD_CONST", 0, 6),
            ("RETURN_VALUE", None, 7)
        ]
        self.cfg_optimization_test(insts, expected, consts=[None])

    def test_format_simple(self):
        # FORMAT_SIMPLE will leave its operand on the stack if it's a unicode
        # object. We treat it conservatively and assume that it always leaves
        # its operand on the stack.
        insts = [
            ("LOAD_FAST", 0, 1),
            ("FORMAT_SIMPLE", None, 2),
            ("STORE_FAST", 1, 3),
        ]
        self.check(insts, insts)

        insts = [
            ("LOAD_FAST", 0, 1),
            ("FORMAT_SIMPLE", None, 2),
            ("POP_TOP", None, 3),
        ]
        expected = [
            ("LOAD_FAST_BORROW", 0, 1),
            ("FORMAT_SIMPLE", None, 2),
            ("POP_TOP", None, 3),
        ]
        self.check(insts, expected)

    def test_set_function_attribute(self):
        # SET_FUNCTION_ATTRIBUTE leaves the function on the stack
        insts = [
            ("LOAD_CONST", 0, 1),
            ("LOAD_FAST", 0, 2),
            ("SET_FUNCTION_ATTRIBUTE", 2, 3),
            ("STORE_FAST", 1, 4),
            ("LOAD_CONST", 0, 5),
            ("RETURN_VALUE", None, 6)
        ]
        self.cfg_optimization_test(insts, insts, consts=[None])

        insts = [
            ("LOAD_CONST", 0, 1),
            ("LOAD_FAST", 0, 2),
            ("SET_FUNCTION_ATTRIBUTE", 2, 3),
            ("RETURN_VALUE", None, 4)
        ]
        expected = [
            ("LOAD_CONST", 0, 1),
            ("LOAD_FAST_BORROW", 0, 2),
            ("SET_FUNCTION_ATTRIBUTE", 2, 3),
            ("RETURN_VALUE", None, 4)
        ]
        self.cfg_optimization_test(insts, expected, consts=[None])

    def test_get_yield_from_iter(self):
        # GET_YIELD_FROM_ITER may leave its operand on the stack
        insts = [
            ("LOAD_FAST", 0, 1),
            ("GET_YIELD_FROM_ITER", None, 2),
            ("LOAD_CONST", 0, 3),
            send := self.Label(),
            ("SEND", end := self.Label(), 5),
            ("YIELD_VALUE", 1, 6),
            ("RESUME", 2, 7),
            ("JUMP", send, 8),
            end,
            ("END_SEND", None, 9),
            ("LOAD_CONST", 0, 10),
            ("RETURN_VALUE", None, 11),
        ]
        self.cfg_optimization_test(insts, insts, consts=[None])

    def test_push_exc_info(self):
        insts = [
            ("LOAD_FAST", 0, 1),
            ("PUSH_EXC_INFO", None, 2),
        ]
        self.check(insts, insts)

    def test_load_special(self):
        # LOAD_SPECIAL may leave self on the stack
        insts = [
            ("LOAD_FAST", 0, 1),
            ("LOAD_SPECIAL", 0, 2),
            ("STORE_FAST", 1, 3),
        ]
        self.check(insts, insts)


    def test_del_in_finally(self):
        # This loads `obj` onto the stack, executes `del obj`, then returns the
        # `obj` from the stack. See gh-133371 for more details.
        def create_obj():
            obj = [42]
            try:
                return obj
            finally:
                del obj

        obj = create_obj()
        # The crash in the linked issue happens while running GC during
        # interpreter finalization, so run it here manually.
        gc.collect()
        self.assertEqual(obj, [42])

    def test_format_simple_unicode(self):
        # Repro from gh-134889
        def f():
            var = f"{1}"
            var = f"{var}"
            return var
        self.assertEqual(f(), "1")



if __name__ == "__main__":
    unittest.main()



"""Tests for C-implemented GenericAlias."""

import unittest
import pickle
from array import array
import copy
from collections import (
    defaultdict, deque, OrderedDict, Counter, UserDict, UserList
)
from collections.abc import *
from concurrent.futures import Future
from concurrent.futures.thread import _WorkItem
from contextlib import AbstractContextManager, AbstractAsyncContextManager
from contextvars import ContextVar, Token
from csv import DictReader, DictWriter
from dataclasses import Field
from functools import partial, partialmethod, cached_property
from graphlib import TopologicalSorter
from logging import LoggerAdapter, StreamHandler
from mailbox import Mailbox, _PartialFile
try:
    import ctypes
except ImportError:
    ctypes = None
from difflib import SequenceMatcher
from filecmp import dircmp
from fileinput import FileInput
from itertools import chain
from http.cookies import Morsel
try:
    from multiprocessing.managers import ValueProxy, DictProxy, ListProxy
    from multiprocessing.pool import ApplyResult
    from multiprocessing.queues import SimpleQueue as MPSimpleQueue
    from multiprocessing.queues import Queue as MPQueue
    from multiprocessing.queues import JoinableQueue as MPJoinableQueue
except ImportError:
    # _multiprocessing module is optional
    ValueProxy = None
    DictProxy = None
    ListProxy = None
    ApplyResult = None
    MPSimpleQueue = None
    MPQueue = None
    MPJoinableQueue = None
try:
    from multiprocessing.shared_memory import ShareableList
except ImportError:
    # multiprocessing.shared_memory is not available on e.g. Android
    ShareableList = None
from os import DirEntry
from re import Pattern, Match
from types import GenericAlias, MappingProxyType, AsyncGeneratorType, CoroutineType, GeneratorType
from tempfile import TemporaryDirectory, SpooledTemporaryFile
from urllib.parse import SplitResult, ParseResult
from unittest.case import _AssertRaisesContext
from queue import Queue, SimpleQueue
from weakref import WeakSet, ReferenceType, ref
import typing
from typing import Unpack
try:
    from tkinter import Event
except ImportError:
    Event = None
from string.templatelib import Template, Interpolation

from typing import TypeVar
T = TypeVar('T')
K = TypeVar('K')
V = TypeVar('V')

_UNPACKED_TUPLES = [
    # Unpacked tuple using `*`
    (*tuple[int],)[0],
    (*tuple[T],)[0],
    (*tuple[int, str],)[0],
    (*tuple[int, ...],)[0],
    (*tuple[T, ...],)[0],
    tuple[*tuple[int, ...]],
    tuple[*tuple[T, ...]],
    tuple[str, *tuple[int, ...]],
    tuple[*tuple[int, ...], str],
    tuple[float, *tuple[int, ...], str],
    tuple[*tuple[*tuple[int, ...]]],
    # Unpacked tuple using `Unpack`
    Unpack[tuple[int]],
    Unpack[tuple[T]],
    Unpack[tuple[int, str]],
    Unpack[tuple[int, ...]],
    Unpack[tuple[T, ...]],
    tuple[Unpack[tuple[int, ...]]],
    tuple[Unpack[tuple[T, ...]]],
    tuple[str, Unpack[tuple[int, ...]]],
    tuple[Unpack[tuple[int, ...]], str],
    tuple[float, Unpack[tuple[int, ...]], str],
    tuple[Unpack[tuple[Unpack[tuple[int, ...]]]]],
    # Unpacked tuple using `*` AND `Unpack`
    tuple[Unpack[tuple[*tuple[int, ...]]]],
    tuple[*tuple[Unpack[tuple[int, ...]]]],
]


class BaseTest(unittest.TestCase):
    """Test basics."""
    generic_types = [type, tuple, list, dict, set, frozenset, enumerate, memoryview,
                     defaultdict, deque,
                     SequenceMatcher,
                     dircmp,
                     FileInput,
                     OrderedDict, Counter, UserDict, UserList,
                     Pattern, Match,
                     partial, partialmethod, cached_property,
                     TopologicalSorter,
                     AbstractContextManager, AbstractAsyncContextManager,
                     Awaitable, Coroutine,
                     AsyncIterable, AsyncIterator,
                     AsyncGenerator, Generator,
                     Iterable, Iterator,
                     Reversible,
                     Container, Collection,
                     Mailbox, _PartialFile,
                     ContextVar, Token,
                     Field,
                     Set, MutableSet,
                     Mapping, MutableMapping, MappingView,
                     KeysView, ItemsView, ValuesView,
                     Sequence, MutableSequence,
                     MappingProxyType, AsyncGeneratorType,
                     GeneratorType, CoroutineType,
                     DirEntry,
                     chain,
                     LoggerAdapter, StreamHandler,
                     TemporaryDirectory, SpooledTemporaryFile,
                     Queue, SimpleQueue,
                     _AssertRaisesContext,
                     SplitResult, ParseResult,
                     WeakSet, ReferenceType, ref,
                     ShareableList,
                     Future, _WorkItem,
                     Morsel,
                     DictReader, DictWriter,
                     array,
                     staticmethod,
                     classmethod,
                     Template,
                     Interpolation,
                    ]
    if ctypes is not None:
        generic_types.extend((ctypes.Array, ctypes.LibraryLoader, ctypes.py_object))
    if ValueProxy is not None:
        generic_types.extend((ValueProxy, DictProxy, ListProxy, ApplyResult,
                              MPSimpleQueue, MPQueue, MPJoinableQueue))
    if Event is not None:
        generic_types.append(Event)

    def test_subscriptable(self):
        for t in self.generic_types:
            if t is None:
                continue
            tname = t.__name__
            with self.subTest(f"Testing {tname}"):
                alias = t[int]
                self.assertIs(alias.__origin__, t)
                self.assertEqual(alias.__args__, (int,))
                self.assertEqual(alias.__parameters__, ())

    def test_unsubscriptable(self):
        for t in int, str, float, Sized, Hashable:
            tname = t.__name__
            with self.subTest(f"Testing {tname}"):
                with self.assertRaisesRegex(TypeError, tname):
                    t[int]

    def test_instantiate(self):
        for t in tuple, list, dict, set, frozenset, defaultdict, deque:
            tname = t.__name__
            with self.subTest(f"Testing {tname}"):
                alias = t[int]
                self.assertEqual(alias(), t())
                if t is dict:
                    self.assertEqual(alias(iter([('a', 1), ('b', 2)])), dict(a=1, b=2))
                    self.assertEqual(alias(a=1, b=2), dict(a=1, b=2))
                elif t is defaultdict:
                    def default():
                        return 'value'
                    a = alias(default)
                    d = defaultdict(default)
                    self.assertEqual(a['test'], d['test'])
                else:
                    self.assertEqual(alias(iter((1, 2, 3))), t((1, 2, 3)))

    def test_unbound_methods(self):
        t = list[int]
        a = t()
        t.append(a, 'foo')
        self.assertEqual(a, ['foo'])
        x = t.__getitem__(a, 0)
        self.assertEqual(x, 'foo')
        self.assertEqual(t.__len__(a), 1)

    def test_subclassing(self):
        class C(list[int]):
            pass
        self.assertEqual(C.__bases__, (list,))
        self.assertEqual(C.__class__, type)

    def test_class_methods(self):
        t = dict[int, None]
        self.assertEqual(dict.fromkeys(range(2)), {0: None, 1: None})  # This works
        self.assertEqual(t.fromkeys(range(2)), {0: None, 1: None})  # Should be equivalent

    def test_no_chaining(self):
        t = list[int]
        with self.assertRaises(TypeError):
            t[int]

    def test_generic_subclass(self):
        class MyList(list):
            pass
        t = MyList[int]
        self.assertIs(t.__origin__, MyList)
        self.assertEqual(t.__args__, (int,))
        self.assertEqual(t.__parameters__, ())

    def test_repr(self):
        class MyList(list):
            pass
        class MyGeneric:
            __class_getitem__ = classmethod(GenericAlias)

        self.assertEqual(repr(list[str]), 'list[str]')
        self.assertEqual(repr(list[()]), 'list[()]')
        self.assertEqual(repr(tuple[int, ...]), 'tuple[int, ...]')
        x1 = tuple[*tuple[int]]
        self.assertEqual(repr(x1), 'tuple[*tuple[int]]')
        x2 = tuple[*tuple[int, str]]
        self.assertEqual(repr(x2), 'tuple[*tuple[int, str]]')
        x3 = tuple[*tuple[int, ...]]
        self.assertEqual(repr(x3), 'tuple[*tuple[int, ...]]')
        self.assertEndsWith(repr(MyList[int]), '.BaseTest.test_repr.<locals>.MyList[int]')
        self.assertEqual(repr(list[str]()), '[]')  # instances should keep their normal repr

        # gh-105488
        self.assertEndsWith(repr(MyGeneric[int]), 'MyGeneric[int]')
        self.assertEndsWith(repr(MyGeneric[[]]), 'MyGeneric[[]]')
        self.assertEndsWith(repr(MyGeneric[[int, str]]), 'MyGeneric[[int, str]]')

    def test_evil_repr1(self):
        # gh-143635
        class Zap:
            def __init__(self, container):
                self.container = container
            def __getattr__(self, name):
                if name == "__origin__":
                    self.container.clear()
                    return None
                if name == "__args__":
                    return ()
                raise AttributeError

        params = []
        params.append(Zap(params))
        alias = GenericAlias(list, (params,))
        repr_str = repr(alias)
        self.assertTrue(repr_str.startswith("list[["), repr_str)

    def test_evil_repr2(self):
        class Zap:
            def __init__(self, container):
                self.container = container
            def __getattr__(self, name):
                if name == "__qualname__":
                    self.container.clear()
                    return "abcd"
                if name == "__module__":
                    return None
                raise AttributeError

        params = []
        params.append(Zap(params))
        alias = GenericAlias(list, (params,))
        repr_str = repr(alias)
        self.assertTrue(repr_str.startswith("list[["), repr_str)

    def test_evil_repr3(self):
        # gh-143823
        lst = []
        class X:
            def __repr__(self):
                lst.clear()
                return "x"

        lst += [X(), 1]
        ga = GenericAlias(int, lst)
        with self.assertRaises(IndexError):
            repr(ga)

    def test_exposed_type(self):
        import types
        a = types.GenericAlias(list, int)
        self.assertEqual(str(a), 'list[int]')
        self.assertIs(a.__origin__, list)
        self.assertEqual(a.__args__, (int,))
        self.assertEqual(a.__parameters__, ())

    def test_parameters(self):
        from typing import List, Dict, Callable

        D0 = dict[str, int]
        self.assertEqual(D0.__args__, (str, int))
        self.assertEqual(D0.__parameters__, ())
        D1a = dict[str, V]
        self.assertEqual(D1a.__args__, (str, V))
        self.assertEqual(D1a.__parameters__, (V,))
        D1b = dict[K, int]
        self.assertEqual(D1b.__args__, (K, int))
        self.assertEqual(D1b.__parameters__, (K,))
        D2a = dict[K, V]
        self.assertEqual(D2a.__args__, (K, V))
        self.assertEqual(D2a.__parameters__, (K, V))
        D2b = dict[T, T]
        self.assertEqual(D2b.__args__, (T, T))
        self.assertEqual(D2b.__parameters__, (T,))

        L0 = list[str]
        self.assertEqual(L0.__args__, (str,))
        self.assertEqual(L0.__parameters__, ())
        L1 = list[T]
        self.assertEqual(L1.__args__, (T,))
        self.assertEqual(L1.__parameters__, (T,))
        L2 = list[list[T]]
        self.assertEqual(L2.__args__, (list[T],))
        self.assertEqual(L2.__parameters__, (T,))
        L3 = list[List[T]]
        self.assertEqual(L3.__args__, (List[T],))
        self.assertEqual(L3.__parameters__, (T,))
        L4a = list[Dict[K, V]]
        self.assertEqual(L4a.__args__, (Dict[K, V],))
        self.assertEqual(L4a.__parameters__, (K, V))
        L4b = list[Dict[T, int]]
        self.assertEqual(L4b.__args__, (Dict[T, int],))
        self.assertEqual(L4b.__parameters__, (T,))
        L5 = list[Callable[[K, V], K]]
        self.assertEqual(L5.__args__, (Callable[[K, V], K],))
        self.assertEqual(L5.__parameters__, (K, V))

        T1 = tuple[*tuple[int]]
        self.assertEqual(
            T1.__args__,
            (*tuple[int],),
        )
        self.assertEqual(T1.__parameters__, ())

        T2 = tuple[*tuple[T]]
        self.assertEqual(
            T2.__args__,
            (*tuple[T],),
        )
        self.assertEqual(T2.__parameters__, (T,))

        T4 = tuple[*tuple[int, str]]
        self.assertEqual(
            T4.__args__,
            (*tuple[int, str],),
        )
        self.assertEqual(T4.__parameters__, ())

    def test_parameter_chaining(self):
        from typing import List, Dict, Union, Callable
        self.assertEqual(list[T][int], list[int])
        self.assertEqual(dict[str, T][int], dict[str, int])
        self.assertEqual(dict[T, int][str], dict[str, int])
        self.assertEqual(dict[K, V][str, int], dict[str, int])
        self.assertEqual(dict[T, T][int], dict[int, int])

        self.assertEqual(list[list[T]][int], list[list[int]])
        self.assertEqual(list[dict[T, int]][str], list[dict[str, int]])
        self.assertEqual(list[dict[str, T]][int], list[dict[str, int]])
        self.assertEqual(list[dict[K, V]][str, int], list[dict[str, int]])
        self.assertEqual(dict[T, list[int]][str], dict[str, list[int]])

        self.assertEqual(list[List[T]][int], list[List[int]])
        self.assertEqual(list[Dict[K, V]][str, int], list[Dict[str, int]])
        self.assertEqual(list[Union[K, V]][str, int], list[Union[str, int]])
        self.assertEqual(list[Callable[[K, V], K]][str, int],
                         list[Callable[[str, int], str]])
        self.assertEqual(dict[T, List[int]][str], dict[str, List[int]])

        with self.assertRaises(TypeError):
            list[int][int]
        with self.assertRaises(TypeError):
            dict[T, int][str, int]
        with self.assertRaises(TypeError):
            dict[str, T][str, int]
        with self.assertRaises(TypeError):
            dict[T, T][str, int]

    def test_equality(self):
        self.assertEqual(list[int], list[int])
        self.assertEqual(dict[str, int], dict[str, int])
        self.assertEqual((*tuple[int],)[0], (*tuple[int],)[0])
        self.assertEqual(tuple[*tuple[int]], tuple[*tuple[int]])
        self.assertNotEqual(dict[str, int], dict[str, str])
        self.assertNotEqual(list, list[int])
        self.assertNotEqual(list[int], list)
        self.assertNotEqual(list[int], tuple[int])
        self.assertNotEqual((*tuple[int],)[0], tuple[int])

    def test_isinstance(self):
        self.assertTrue(isinstance([], list))
        with self.assertRaises(TypeError):
            isinstance([], list[str])

    def test_issubclass(self):
        class L(list): ...
        self.assertIsSubclass(L, list)
        with self.assertRaises(TypeError):
            issubclass(L, list[str])

    def test_type_generic(self):
        t = type[int]
        Test = t('Test', (), {})
        self.assertTrue(isinstance(Test, type))
        test = Test()
        self.assertEqual(t(test), Test)
        self.assertEqual(t(0), int)

    def test_type_subclass_generic(self):
        class MyType(type):
            pass
        with self.assertRaisesRegex(TypeError, 'MyType'):
            MyType[int]

    def test_pickle(self):
        aliases = [GenericAlias(list, T)] + _UNPACKED_TUPLES
        for alias in aliases:
            for proto in range(pickle.HIGHEST_PROTOCOL + 1):
                with self.subTest(alias=alias, proto=proto):
                    s = pickle.dumps(alias, proto)
                    loaded = pickle.loads(s)
                    self.assertEqual(loaded.__origin__, alias.__origin__)
                    self.assertEqual(loaded.__args__, alias.__args__)
                    self.assertEqual(loaded.__parameters__, alias.__parameters__)
                    self.assertEqual(type(loaded), type(alias))

    def test_copy(self):
        class X(list):
            def __copy__(self):
                return self
            def __deepcopy__(self, memo):
                return self

        aliases = [
            GenericAlias(list, T),
            GenericAlias(deque, T),
            GenericAlias(X, T),
            X[T],
            list[T],
            deque[T],
        ] + _UNPACKED_TUPLES
        for alias in aliases:
            with self.subTest(alias=alias):
                copied = copy.copy(alias)
                self.assertEqual(copied.__origin__, alias.__origin__)
                self.assertEqual(copied.__args__, alias.__args__)
                self.assertEqual(copied.__parameters__, alias.__parameters__)
                copied = copy.deepcopy(alias)
                self.assertEqual(copied.__origin__, alias.__origin__)
                self.assertEqual(copied.__args__, alias.__args__)
                self.assertEqual(copied.__parameters__, alias.__parameters__)

    def test_unpack(self):
        alias = tuple[str, ...]
        self.assertIs(alias.__unpacked__, False)
        unpacked = (*alias,)[0]
        self.assertIs(unpacked.__unpacked__, True)

    def test_union(self):
        a = typing.Union[list[int], list[str]]
        self.assertEqual(a.__args__, (list[int], list[str]))
        self.assertEqual(a.__parameters__, ())

    def test_union_generic(self):
        a = typing.Union[list[T], tuple[T, ...]]
        self.assertEqual(a.__args__, (list[T], tuple[T, ...]))
        self.assertEqual(a.__parameters__, (T,))

    def test_dir(self):
        ga = list[int]
        dir_of_gen_alias = set(dir(ga))
        self.assertTrue(dir_of_gen_alias.issuperset(dir(list)))
        for generic_alias_property in (
            "__origin__", "__args__", "__parameters__",
            "__unpacked__",
        ):
            with self.subTest(generic_alias_property=generic_alias_property):
                self.assertIn(generic_alias_property, dir_of_gen_alias)
        for blocked in (
            "__bases__",
            "__copy__",
            "__deepcopy__",
        ):
            with self.subTest(blocked=blocked):
                self.assertNotIn(blocked, dir_of_gen_alias)

        for entry in dir_of_gen_alias:
            with self.subTest(entry=entry):
                getattr(ga, entry)  # must not raise `AttributeError`

    def test_weakref(self):
        for t in self.generic_types:
            if t is None:
                continue
            tname = t.__name__
            with self.subTest(f"Testing {tname}"):
                alias = t[int]
                self.assertEqual(ref(alias)(), alias)

    def test_no_kwargs(self):
        # bpo-42576
        with self.assertRaises(TypeError):
            GenericAlias(bad=float)

    def test_subclassing_types_genericalias(self):
        class SubClass(GenericAlias): ...
        alias = SubClass(list, int)
        class Bad(GenericAlias):
            def __new__(cls, *args, **kwargs):
                super().__new__(cls, *args, **kwargs)

        self.assertEqual(alias, list[int])
        with self.assertRaises(TypeError):
            Bad(list, int, bad=int)

    def test_iter_creates_starred_tuple(self):
        t = tuple[int, str]
        iter_t = iter(t)
        x = next(iter_t)
        self.assertEqual(repr(x), '*tuple[int, str]')

    def test_calling_next_twice_raises_stopiteration(self):
        t = tuple[int, str]
        iter_t = iter(t)
        next(iter_t)
        with self.assertRaises(StopIteration):
            next(iter_t)

    def test_del_iter(self):
        t = tuple[int, str]
        iter_x = iter(t)
        del iter_x

    def test_paramspec_specialization(self):
        # gh-124445
        T = TypeVar("T")
        U = TypeVar("U")
        type X[**P] = Callable[P, int]

        generic = X[[T]]
        self.assertEqual(generic.__args__, ([T],))
        self.assertEqual(generic.__parameters__, (T,))
        specialized = generic[str]
        self.assertEqual(specialized.__args__, ([str],))
        self.assertEqual(specialized.__parameters__, ())

        generic = X[(T,)]
        self.assertEqual(generic.__args__, (T,))
        self.assertEqual(generic.__parameters__, (T,))
        specialized = generic[str]
        self.assertEqual(specialized.__args__, (str,))
        self.assertEqual(specialized.__parameters__, ())

        generic = X[[T, U]]
        self.assertEqual(generic.__args__, ([T, U],))
        self.assertEqual(generic.__parameters__, (T, U))
        specialized = generic[str, int]
        self.assertEqual(specialized.__args__, ([str, int],))
        self.assertEqual(specialized.__parameters__, ())

        generic = X[(T, U)]
        self.assertEqual(generic.__args__, (T, U))
        self.assertEqual(generic.__parameters__, (T, U))
        specialized = generic[str, int]
        self.assertEqual(specialized.__args__, (str, int))
        self.assertEqual(specialized.__parameters__, ())

    def test_nested_paramspec_specialization(self):
        # gh-124445
        type X[**P, T] = Callable[P, T]

        x_list = X[[int, str], float]
        self.assertEqual(x_list.__args__, ([int, str], float))
        self.assertEqual(x_list.__parameters__, ())

        x_tuple = X[(int, str), float]
        self.assertEqual(x_tuple.__args__, ((int, str), float))
        self.assertEqual(x_tuple.__parameters__, ())

        U = TypeVar("U")
        V = TypeVar("V")

        multiple_params_list = X[[int, U], V]
        self.assertEqual(multiple_params_list.__args__, ([int, U], V))
        self.assertEqual(multiple_params_list.__parameters__, (U, V))
        multiple_params_list_specialized = multiple_params_list[str, float]
        self.assertEqual(multiple_params_list_specialized.__args__, ([int, str], float))
        self.assertEqual(multiple_params_list_specialized.__parameters__, ())

        multiple_params_tuple = X[(int, U), V]
        self.assertEqual(multiple_params_tuple.__args__, ((int, U), V))
        self.assertEqual(multiple_params_tuple.__parameters__, (U, V))
        multiple_params_tuple_specialized = multiple_params_tuple[str, float]
        self.assertEqual(multiple_params_tuple_specialized.__args__, ((int, str), float))
        self.assertEqual(multiple_params_tuple_specialized.__parameters__, ())

        deeply_nested = X[[U, [V], int], V]
        self.assertEqual(deeply_nested.__args__, ([U, [V], int], V))
        self.assertEqual(deeply_nested.__parameters__, (U, V))
        deeply_nested_specialized = deeply_nested[str, float]
        self.assertEqual(deeply_nested_specialized.__args__, ([str, [float], int], float))
        self.assertEqual(deeply_nested_specialized.__parameters__, ())


class TypeIterationTests(unittest.TestCase):
    _UNITERABLE_TYPES = (list, tuple)

    def test_cannot_iterate(self):
        for test_type in self._UNITERABLE_TYPES:
            with self.subTest(type=test_type):
                expected_error_regex = "object is not iterable"
                with self.assertRaisesRegex(TypeError, expected_error_regex):
                    iter(test_type)
                with self.assertRaisesRegex(TypeError, expected_error_regex):
                    list(test_type)
                with self.assertRaisesRegex(TypeError, expected_error_regex):
                    for _ in test_type:
                        pass

    def test_is_not_instance_of_iterable(self):
        for type_to_test in self._UNITERABLE_TYPES:
            self.assertNotIsInstance(type_to_test, Iterable)


if __name__ == "__main__":
    unittest.main()



import unittest

from threading import Thread, Barrier
from unittest import TestCase

from test.support import threading_helper


@threading_helper.requires_working_threading()
class TestSet(TestCase):
    def test_repr_clear(self):
        """Test repr() of a set while another thread is calling clear()"""
        NUM_ITERS = 10
        NUM_REPR_THREADS = 10
        barrier = Barrier(NUM_REPR_THREADS + 1)
        s = {1, 2, 3, 4, 5, 6, 7, 8}

        def clear_set():
            barrier.wait()
            s.clear()

        def repr_set():
            barrier.wait()
            set_reprs.append(repr(s))

        for _ in range(NUM_ITERS):
            set_reprs = []
            threads = [Thread(target=clear_set)]
            for _ in range(NUM_REPR_THREADS):
                threads.append(Thread(target=repr_set))
            for t in threads:
                t.start()
            for t in threads:
                t.join()

            for set_repr in set_reprs:
                self.assertIn(set_repr, ("set()", "{1, 2, 3, 4, 5, 6, 7, 8}"))


if __name__ == "__main__":
    unittest.main()



import unittest
import tkinter
from test.support import requires, swap_attr
from test.test_tkinter.support import setUpModule  # noqa: F401
from test.test_tkinter.support import AbstractDefaultRootTest
from tkinter.commondialog import Dialog
from tkinter.messagebox import showinfo

requires('gui')


class DefaultRootTest(AbstractDefaultRootTest, unittest.TestCase):

    def test_showinfo(self):
        def test_callback(dialog, master):
            nonlocal ismapped
            master.update()
            ismapped = master.winfo_ismapped()
            raise ZeroDivisionError

        with swap_attr(Dialog, '_test_callback', test_callback):
            ismapped = None
            self.assertRaises(ZeroDivisionError, showinfo, "Spam", "Egg Information")
            self.assertEqual(ismapped, False)

            root = tkinter.Tk()
            ismapped = None
            self.assertRaises(ZeroDivisionError, showinfo, "Spam", "Egg Information")
            self.assertEqual(ismapped, True)
            root.destroy()

            tkinter.NoDefaultRoot()
            self.assertRaises(RuntimeError, showinfo, "Spam", "Egg Information")


if __name__ == "__main__":
    unittest.main()



"""This test checks for correct wait4() behavior.
"""

import os
import sys
import unittest
from test.fork_wait import ForkWait
from test import support

# If either of these do not exist, skip this test.
if not support.has_fork_support:
    raise unittest.SkipTest("requires working os.fork()")

support.get_attribute(os, 'wait4')


class Wait4Test(ForkWait):
    def wait_impl(self, cpid, *, exitcode):
        option = os.WNOHANG
        if sys.platform.startswith('aix'):
            # Issue #11185: wait4 is broken on AIX and will always return 0
            # with WNOHANG.
            option = 0
        for _ in support.sleeping_retry(support.SHORT_TIMEOUT):
            # wait4() shouldn't hang, but some of the buildbots seem to hang
            # in the forking tests.  This is an attempt to fix the problem.
            spid, status, rusage = os.wait4(cpid, option)
            if spid == cpid:
                break
        self.assertEqual(spid, cpid)
        self.assertEqual(os.waitstatus_to_exitcode(status), exitcode)
        self.assertTrue(rusage)

def tearDownModule():
    support.reap_children()

if __name__ == "__main__":
    unittest.main()
